import	os
import	unittest
from	time								import sleep
from	logging								import Logger
from	logging								import StreamHandler
from	logging								import FileHandler
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.magical.time_turner.timers	import DIRTtimer
from	pygwarts.tests.irma					import IrmaTestCase
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.irma.contrib.intercept		import ContribInterceptor
from	pygwarts.irma.contrib.intercept		import PoolHoist








class ContribInterceptingCase(IrmaTestCase):

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.INTERCEPT_HANDLER): os.remove(cls.INTERCEPT_HANDLER)


	class DebugAugmentor(ContribInterceptor):

		def __call__(self):
			class Interceptor(self.layer):

				def debug(self, message :str): return super().debug(f"{message} (augmented)")
			return	Interceptor

	class InfoAugmentor(ContribInterceptor):

		def __call__(self):
			class Interceptor(self.layer):

				def info(self, message :str): return super().info(f"{message} (augmented)")
			return	Interceptor

	class WarningAugmentor(ContribInterceptor):

		def __call__(self):
			class Interceptor(self.layer):

				def warning(self, message :str): return super().warning(f"{message} (augmented)")
			return	Interceptor

	class ErrorAugmentor(ContribInterceptor):

		def __call__(self):
			class Interceptor(self.layer):

				def error(self, message :str): return super().error(f"{message} (augmented)")
			return	Interceptor

	class CriticalAugmentor(ContribInterceptor):

		def __call__(self):
			class Interceptor(self.layer):

				def critical(self, message :str): return super().critical(f"{message} (augmented)")
			return	Interceptor

	class InfoHoist(ContribInterceptor):

		def __call__(self):
			class Interceptor(self.layer):

				def info(self, message :str):
					logged = super().info(message)
					if hasattr(self, "buffer_insert"): self.buffer_insert(message)
					return logged
			return	Interceptor

	class ReleaseWriter1(ContribInterceptor):
		def __call__(self):

			class Interceptor(self.layer):
				def buffer_release(self, *args, **kwargs):

					buffer_dump = super().buffer_release(*args, **kwargs)
					with open(

						IrmaTestCase.IRMA_ROOT /"ih2w1.release",
						"w"
					)	as	writer:
						for line in buffer_dump : writer.write(f"{line}\n")
			return	Interceptor

	class ReleaseWriter2(PoolHoist):
		def __call__(self):

			class Interceptor(super().__call__()):
				def buffer_release(self, *args, **kwargs):

					buffer_dump = super().buffer_release(*args, **kwargs)
					with open(

						IrmaTestCase.IRMA_ROOT /"ih2w2.release",
						"w"
					)	as	writer:
						for line in buffer_dump : writer.write(f"{line}\n")
			return	Interceptor


	def test_PoolHoist_default_members(self):
		class Irma(Transmutable):

			@PoolHoist
			class loggy(LibraryContrib): pass


		self.test_case = Irma()
		self.assertTrue(hasattr(self.test_case.loggy, "handler"))
		self.assertIsInstance(self.test_case.loggy.handler, StreamHandler)
		self.assertTrue(hasattr(self.test_case.loggy, "init_name"))
		self.assertEqual(self.test_case.loggy.init_name, "fantastic logs and where to contribute them")
		self.assertTrue(hasattr(self.test_case.loggy, "init_level"))
		self.assertEqual(self.test_case.loggy.init_level, 20)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_mode"))
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(hasattr(self.test_case.loggy, "contribfmt"))
		self.assertEqual(

			self.test_case.loggy.contribfmt,
			{
				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s : %(message)s",
				"datefmt": "%d/%m/%Y %H%M",
			}
		)

		self.assertTrue(hasattr(self.test_case.loggy, "forcing_map"))
		self.assertIsInstance(self.test_case.loggy.forcing_map, dict)
		self.assertFalse(self.test_case.loggy.forcing_map)

		self.assertTrue(hasattr(self.test_case.loggy, "handover_map"))
		self.assertIsInstance(self.test_case.loggy.handover_map, dict)
		self.assertEqual(self.test_case.loggy.handover_map, { "Irma": 20 })

		self.assertTrue(hasattr(self.test_case.loggy, "contributor_name"))
		self.assertIsNone(self.test_case.loggy.contributor_name)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_order"))
		self.assertEqual(self.test_case.loggy.handover_order, [ "Irma" ])
		self.assertTrue(hasattr(self.test_case.loggy, "contributor"))
		self.assertIsInstance(self.test_case.loggy.contributor, Logger)

		self.assertTrue(hasattr(self.test_case.loggy, "POOL_TIMER"))
		self.assertEqual(self.test_case.loggy.POOL_TIMER, 5)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_LIMIT"))
		self.assertEqual(self.test_case.loggy.POOL_LIMIT, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_NAME"))
		self.assertIsInstance(self.test_case.loggy.POOL_NAME, str)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_STATE"))
		self.assertEqual(self.test_case.loggy.POOL_STATE, 0)
		self.assertTrue(hasattr(self.test_case.loggy, "BUFFER"))
		self.assertEqual(self.test_case.loggy.BUFFER, [])








	def test_PoolHoist_name_default_members(self):
		class Irma(Transmutable):

			@PoolHoist
			class shmoggy(LibraryContrib): pass


		self.test_case = Irma()
		self.assertTrue(hasattr(self.test_case.loggy, "handler"))
		self.assertIsInstance(self.test_case.loggy.handler, StreamHandler)
		self.assertTrue(hasattr(self.test_case.loggy, "init_name"))
		self.assertEqual(self.test_case.loggy.init_name, "fantastic logs and where to contribute them")
		self.assertTrue(hasattr(self.test_case.loggy, "init_level"))
		self.assertEqual(self.test_case.loggy.init_level, 20)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_mode"))
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(hasattr(self.test_case.loggy, "contribfmt"))
		self.assertEqual(

			self.test_case.loggy.contribfmt,
			{
				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s : %(message)s",
				"datefmt": "%d/%m/%Y %H%M",
			}
		)

		self.assertTrue(hasattr(self.test_case.loggy, "forcing_map"))
		self.assertIsInstance(self.test_case.loggy.forcing_map, dict)
		self.assertFalse(self.test_case.loggy.forcing_map)

		self.assertTrue(hasattr(self.test_case.loggy, "handover_map"))
		self.assertIsInstance(self.test_case.loggy.handover_map, dict)
		self.assertEqual(self.test_case.loggy.handover_map, { "Irma": 20 })

		self.assertTrue(hasattr(self.test_case.loggy, "contributor_name"))
		self.assertIsNone(self.test_case.loggy.contributor_name)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_order"))
		self.assertEqual(self.test_case.loggy.handover_order, [ "Irma" ])
		self.assertTrue(hasattr(self.test_case.loggy, "contributor"))
		self.assertIsInstance(self.test_case.loggy.contributor, Logger)

		self.assertTrue(hasattr(self.test_case.loggy, "POOL_TIMER"))
		self.assertEqual(self.test_case.loggy.POOL_TIMER, 5)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_LIMIT"))
		self.assertEqual(self.test_case.loggy.POOL_LIMIT, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_NAME"))
		self.assertIsInstance(self.test_case.loggy.POOL_NAME, str)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_STATE"))
		self.assertEqual(self.test_case.loggy.POOL_STATE, 0)
		self.assertTrue(hasattr(self.test_case.loggy, "BUFFER"))
		self.assertEqual(self.test_case.loggy.BUFFER, [])








	def test_PoolHoist_escalated_default_members(self):

		class Irma(Transmutable):
			class Shirma(Transmutable):

				@PoolHoist
				class loggy(LibraryContrib): pass


		self.test_case = Irma()
		self.assertTrue(hasattr(self.test_case.loggy, "handler"))
		self.assertIsInstance(self.test_case.loggy.handler, StreamHandler)
		self.assertTrue(hasattr(self.test_case.loggy, "init_name"))
		self.assertEqual(self.test_case.loggy.init_name, "fantastic logs and where to contribute them")
		self.assertTrue(hasattr(self.test_case.loggy, "init_level"))
		self.assertEqual(self.test_case.loggy.init_level, 20)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_mode"))
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(hasattr(self.test_case.loggy, "contribfmt"))
		self.assertEqual(

			self.test_case.loggy.contribfmt,
			{
				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s : %(message)s",
				"datefmt": "%d/%m/%Y %H%M",
			}
		)

		self.assertTrue(hasattr(self.test_case.loggy, "forcing_map"))
		self.assertIsInstance(self.test_case.loggy.forcing_map, dict)
		self.assertFalse(self.test_case.loggy.forcing_map)

		self.assertTrue(hasattr(self.test_case.loggy, "handover_map"))
		self.assertIsInstance(self.test_case.loggy.handover_map, dict)
		self.assertEqual(self.test_case.loggy.handover_map, { "Irma": 20, "Irma.Shirma": 20 })

		self.assertTrue(hasattr(self.test_case.loggy, "contributor_name"))
		self.assertIsNone(self.test_case.loggy.contributor_name)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_order"))
		self.assertEqual(self.test_case.loggy.handover_order, [ "Irma.Shirma", "Irma" ])
		self.assertTrue(hasattr(self.test_case.loggy, "contributor"))
		self.assertIsInstance(self.test_case.loggy.contributor, Logger)

		self.assertTrue(hasattr(self.test_case.loggy, "POOL_TIMER"))
		self.assertEqual(self.test_case.loggy.POOL_TIMER, 5)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_LIMIT"))
		self.assertEqual(self.test_case.loggy.POOL_LIMIT, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_NAME"))
		self.assertIsInstance(self.test_case.loggy.POOL_NAME, str)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_STATE"))
		self.assertEqual(self.test_case.loggy.POOL_STATE, 0)
		self.assertTrue(hasattr(self.test_case.loggy, "BUFFER"))
		self.assertEqual(self.test_case.loggy.BUFFER, [])








	def test_PoolHoist_escalated_name_default_members(self):

		class Irma(Transmutable):
			class Shirma(Transmutable):

				@PoolHoist
				class shmoggy(LibraryContrib): pass


		self.test_case = Irma()
		self.assertTrue(hasattr(self.test_case.loggy, "handler"))
		self.assertIsInstance(self.test_case.loggy.handler, StreamHandler)
		self.assertTrue(hasattr(self.test_case.loggy, "init_name"))
		self.assertEqual(self.test_case.loggy.init_name, "fantastic logs and where to contribute them")
		self.assertTrue(hasattr(self.test_case.loggy, "init_level"))
		self.assertEqual(self.test_case.loggy.init_level, 20)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_mode"))
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(hasattr(self.test_case.loggy, "contribfmt"))
		self.assertEqual(

			self.test_case.loggy.contribfmt,
			{
				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s : %(message)s",
				"datefmt": "%d/%m/%Y %H%M",
			}
		)

		self.assertTrue(hasattr(self.test_case.loggy, "forcing_map"))
		self.assertIsInstance(self.test_case.loggy.forcing_map, dict)
		self.assertFalse(self.test_case.loggy.forcing_map)

		self.assertTrue(hasattr(self.test_case.loggy, "handover_map"))
		self.assertIsInstance(self.test_case.loggy.handover_map, dict)
		self.assertEqual(self.test_case.loggy.handover_map, { "Irma": 20, "Irma.Shirma": 20 })

		self.assertTrue(hasattr(self.test_case.loggy, "contributor_name"))
		self.assertIsNone(self.test_case.loggy.contributor_name)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_order"))
		self.assertEqual(self.test_case.loggy.handover_order, [ "Irma.Shirma", "Irma" ])
		self.assertTrue(hasattr(self.test_case.loggy, "contributor"))
		self.assertIsInstance(self.test_case.loggy.contributor, Logger)

		self.assertTrue(hasattr(self.test_case.loggy, "POOL_TIMER"))
		self.assertEqual(self.test_case.loggy.POOL_TIMER, 5)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_LIMIT"))
		self.assertEqual(self.test_case.loggy.POOL_LIMIT, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_NAME"))
		self.assertIsInstance(self.test_case.loggy.POOL_NAME, str)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_STATE"))
		self.assertEqual(self.test_case.loggy.POOL_STATE, 0)
		self.assertTrue(hasattr(self.test_case.loggy, "BUFFER"))
		self.assertEqual(self.test_case.loggy.BUFFER, [])








	def test_PoolHoist_field_members(self):
		class Irma(Transmutable):

			@PoolHoist
			class loggy(LibraryContrib):

				handler			= self.INTERCEPT_HANDLER
				init_name		= "field_members"
				init_level		= 50
				force_handover	= False
				force_debug		= []
				force_info		= []
				force_warning	= []
				force_error		= []
				force_critical	= []
				contribfmt		= {

					"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s - %(message)s",
					"datefmt": "%Y/%d/%m %H%M",
				}


		self.test_case = Irma()
		self.assertTrue(hasattr(self.test_case.loggy, "handler"))
		self.assertIsInstance(self.test_case.loggy.handler, FileHandler)
		self.assertTrue(hasattr(self.test_case.loggy, "init_name"))
		self.assertEqual(self.test_case.loggy.init_name, "field_members")
		self.assertTrue(hasattr(self.test_case.loggy, "init_level"))
		self.assertEqual(self.test_case.loggy.init_level, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_mode"))
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(hasattr(self.test_case.loggy, "contribfmt"))
		self.assertEqual(

			self.test_case.loggy.contribfmt,
			{
				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s - %(message)s",
				"datefmt": "%Y/%d/%m %H%M",
			}
		)

		self.assertTrue(hasattr(self.test_case.loggy, "forcing_map"))
		self.assertIsInstance(self.test_case.loggy.forcing_map, dict)
		self.assertFalse(self.test_case.loggy.forcing_map)

		self.assertTrue(hasattr(self.test_case.loggy, "handover_map"))
		self.assertIsInstance(self.test_case.loggy.handover_map, dict)
		self.assertEqual(self.test_case.loggy.handover_map, { "Irma": 50 })

		self.assertTrue(hasattr(self.test_case.loggy, "contributor_name"))
		self.assertIsNone(self.test_case.loggy.contributor_name)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_order"))
		self.assertEqual(self.test_case.loggy.handover_order, [ "Irma" ])
		self.assertTrue(hasattr(self.test_case.loggy, "contributor"))
		self.assertIsInstance(self.test_case.loggy.contributor, Logger)

		self.assertTrue(hasattr(self.test_case.loggy, "POOL_TIMER"))
		self.assertEqual(self.test_case.loggy.POOL_TIMER, 5)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_LIMIT"))
		self.assertEqual(self.test_case.loggy.POOL_LIMIT, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_NAME"))
		self.assertIsInstance(self.test_case.loggy.POOL_NAME, str)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_STATE"))
		self.assertEqual(self.test_case.loggy.POOL_STATE, 0)
		self.assertTrue(hasattr(self.test_case.loggy, "BUFFER"))
		self.assertEqual(self.test_case.loggy.BUFFER, [])








	def test_PoolHoist_name_field_members(self):
		class Irma(Transmutable):

			@PoolHoist
			class shmoggy(LibraryContrib):

				handler			= self.INTERCEPT_HANDLER
				init_name		= "field_members"
				init_level		= 50
				force_handover	= False
				force_debug		= []
				force_info		= []
				force_warning	= []
				force_error		= []
				force_critical	= []
				contribfmt		= {

					"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s - %(message)s",
					"datefmt": "%Y/%d/%m %H%M",
				}


		self.test_case = Irma()
		self.assertTrue(hasattr(self.test_case.loggy, "handler"))
		self.assertIsInstance(self.test_case.loggy.handler, FileHandler)
		self.assertTrue(hasattr(self.test_case.loggy, "init_name"))
		self.assertEqual(self.test_case.loggy.init_name, "field_members")
		self.assertTrue(hasattr(self.test_case.loggy, "init_level"))
		self.assertEqual(self.test_case.loggy.init_level, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_mode"))
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(hasattr(self.test_case.loggy, "contribfmt"))
		self.assertEqual(

			self.test_case.loggy.contribfmt,
			{
				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s - %(message)s",
				"datefmt": "%Y/%d/%m %H%M",
			}
		)

		self.assertTrue(hasattr(self.test_case.loggy, "forcing_map"))
		self.assertIsInstance(self.test_case.loggy.forcing_map, dict)
		self.assertFalse(self.test_case.loggy.forcing_map)

		self.assertTrue(hasattr(self.test_case.loggy, "handover_map"))
		self.assertIsInstance(self.test_case.loggy.handover_map, dict)
		self.assertEqual(self.test_case.loggy.handover_map, { "Irma": 50 })

		self.assertTrue(hasattr(self.test_case.loggy, "contributor_name"))
		self.assertIsNone(self.test_case.loggy.contributor_name)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_order"))
		self.assertEqual(self.test_case.loggy.handover_order, [ "Irma" ])
		self.assertTrue(hasattr(self.test_case.loggy, "contributor"))
		self.assertIsInstance(self.test_case.loggy.contributor, Logger)

		self.assertTrue(hasattr(self.test_case.loggy, "POOL_TIMER"))
		self.assertEqual(self.test_case.loggy.POOL_TIMER, 5)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_LIMIT"))
		self.assertEqual(self.test_case.loggy.POOL_LIMIT, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_NAME"))
		self.assertIsInstance(self.test_case.loggy.POOL_NAME, str)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_STATE"))
		self.assertEqual(self.test_case.loggy.POOL_STATE, 0)
		self.assertTrue(hasattr(self.test_case.loggy, "BUFFER"))
		self.assertEqual(self.test_case.loggy.BUFFER, [])








	def test_PoolHoist_escalated_field_members(self):

		class Irma(Transmutable):
			class Shirma(Transmutable):

				@PoolHoist
				class loggy(LibraryContrib):

					handler			= self.INTERCEPT_HANDLER
					init_name		= "field_members"
					init_level		= 50
					force_handover	= False
					force_debug		= []
					force_info		= []
					force_warning	= []
					force_error		= []
					force_critical	= []
					contribfmt		= {

						"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s - %(message)s",
						"datefmt": "%Y/%d/%m %H%M",
					}


		self.test_case = Irma()
		self.assertTrue(hasattr(self.test_case.loggy, "handler"))
		self.assertIsInstance(self.test_case.loggy.handler, FileHandler)
		self.assertTrue(hasattr(self.test_case.loggy, "init_name"))
		self.assertEqual(self.test_case.loggy.init_name, "field_members")
		self.assertTrue(hasattr(self.test_case.loggy, "init_level"))
		self.assertEqual(self.test_case.loggy.init_level, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_mode"))
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(hasattr(self.test_case.loggy, "contribfmt"))
		self.assertEqual(

			self.test_case.loggy.contribfmt,
			{
				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s - %(message)s",
				"datefmt": "%Y/%d/%m %H%M",
			}
		)

		self.assertTrue(hasattr(self.test_case.loggy, "forcing_map"))
		self.assertIsInstance(self.test_case.loggy.forcing_map, dict)
		self.assertFalse(self.test_case.loggy.forcing_map)

		self.assertTrue(hasattr(self.test_case.loggy, "handover_map"))
		self.assertIsInstance(self.test_case.loggy.handover_map, dict)
		self.assertEqual(self.test_case.loggy.handover_map, { "Irma": 50, "Irma.Shirma": 50 })

		self.assertTrue(hasattr(self.test_case.loggy, "contributor_name"))
		self.assertIsNone(self.test_case.loggy.contributor_name)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_order"))
		self.assertEqual(self.test_case.loggy.handover_order, [ "Irma.Shirma", "Irma" ])
		self.assertTrue(hasattr(self.test_case.loggy, "contributor"))
		self.assertIsInstance(self.test_case.loggy.contributor, Logger)

		self.assertTrue(hasattr(self.test_case.loggy, "POOL_TIMER"))
		self.assertEqual(self.test_case.loggy.POOL_TIMER, 5)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_LIMIT"))
		self.assertEqual(self.test_case.loggy.POOL_LIMIT, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_NAME"))
		self.assertIsInstance(self.test_case.loggy.POOL_NAME, str)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_STATE"))
		self.assertEqual(self.test_case.loggy.POOL_STATE, 0)
		self.assertTrue(hasattr(self.test_case.loggy, "BUFFER"))
		self.assertEqual(self.test_case.loggy.BUFFER, [])








	def test_PoolHoist_escalated_name_field_members(self):

		class Irma(Transmutable):
			class Shirma(Transmutable):

				@PoolHoist
				class shmoggy(LibraryContrib):

					handler			= self.INTERCEPT_HANDLER
					init_name		= "field_members"
					init_level		= 50
					force_handover	= False
					force_debug		= []
					force_info		= []
					force_warning	= []
					force_error		= []
					force_critical	= []
					contribfmt		= {

						"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s - %(message)s",
						"datefmt": "%Y/%d/%m %H%M",
					}


		self.test_case = Irma()
		self.assertTrue(hasattr(self.test_case.loggy, "handler"))
		self.assertIsInstance(self.test_case.loggy.handler, FileHandler)
		self.assertTrue(hasattr(self.test_case.loggy, "init_name"))
		self.assertEqual(self.test_case.loggy.init_name, "field_members")
		self.assertTrue(hasattr(self.test_case.loggy, "init_level"))
		self.assertEqual(self.test_case.loggy.init_level, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_mode"))
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(hasattr(self.test_case.loggy, "contribfmt"))
		self.assertEqual(

			self.test_case.loggy.contribfmt,
			{
				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s - %(message)s",
				"datefmt": "%Y/%d/%m %H%M",
			}
		)

		self.assertTrue(hasattr(self.test_case.loggy, "forcing_map"))
		self.assertIsInstance(self.test_case.loggy.forcing_map, dict)
		self.assertFalse(self.test_case.loggy.forcing_map)

		self.assertTrue(hasattr(self.test_case.loggy, "handover_map"))
		self.assertIsInstance(self.test_case.loggy.handover_map, dict)
		self.assertEqual(self.test_case.loggy.handover_map, { "Irma": 50, "Irma.Shirma": 50 })

		self.assertTrue(hasattr(self.test_case.loggy, "contributor_name"))
		self.assertIsNone(self.test_case.loggy.contributor_name)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_order"))
		self.assertEqual(self.test_case.loggy.handover_order, [ "Irma.Shirma", "Irma" ])
		self.assertTrue(hasattr(self.test_case.loggy, "contributor"))
		self.assertIsInstance(self.test_case.loggy.contributor, Logger)

		self.assertTrue(hasattr(self.test_case.loggy, "POOL_TIMER"))
		self.assertEqual(self.test_case.loggy.POOL_TIMER, 5)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_LIMIT"))
		self.assertEqual(self.test_case.loggy.POOL_LIMIT, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_NAME"))
		self.assertIsInstance(self.test_case.loggy.POOL_NAME, str)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_STATE"))
		self.assertEqual(self.test_case.loggy.POOL_STATE, 0)
		self.assertTrue(hasattr(self.test_case.loggy, "BUFFER"))
		self.assertEqual(self.test_case.loggy.BUFFER, [])








	def test_outer_PoolHoist_default_members(self):
		class Outdoor(Transmutable):

			@PoolHoist
			class loggy(LibraryContrib):	pass
		class Irma(Outdoor):				pass


		self.test_case = Irma()
		self.assertTrue(hasattr(self.test_case.loggy, "handler"))
		self.assertIsInstance(self.test_case.loggy.handler, StreamHandler)
		self.assertTrue(hasattr(self.test_case.loggy, "init_name"))
		self.assertEqual(self.test_case.loggy.init_name, "fantastic logs and where to contribute them")
		self.assertTrue(hasattr(self.test_case.loggy, "init_level"))
		self.assertEqual(self.test_case.loggy.init_level, 20)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_mode"))
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(hasattr(self.test_case.loggy, "contribfmt"))
		self.assertEqual(

			self.test_case.loggy.contribfmt,
			{
				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s : %(message)s",
				"datefmt": "%d/%m/%Y %H%M",
			}
		)

		self.assertTrue(hasattr(self.test_case.loggy, "forcing_map"))
		self.assertIsInstance(self.test_case.loggy.forcing_map, dict)
		self.assertFalse(self.test_case.loggy.forcing_map)

		self.assertTrue(hasattr(self.test_case.loggy, "handover_map"))
		self.assertIsInstance(self.test_case.loggy.handover_map, dict)
		self.assertEqual(self.test_case.loggy.handover_map, { "Irma": 20 })

		self.assertTrue(hasattr(self.test_case.loggy, "contributor_name"))
		self.assertIsNone(self.test_case.loggy.contributor_name)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_order"))
		self.assertEqual(self.test_case.loggy.handover_order, [ "Irma" ])
		self.assertTrue(hasattr(self.test_case.loggy, "contributor"))
		self.assertIsInstance(self.test_case.loggy.contributor, Logger)

		self.assertTrue(hasattr(self.test_case.loggy, "POOL_TIMER"))
		self.assertEqual(self.test_case.loggy.POOL_TIMER, 5)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_LIMIT"))
		self.assertEqual(self.test_case.loggy.POOL_LIMIT, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_NAME"))
		self.assertIsInstance(self.test_case.loggy.POOL_NAME, str)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_STATE"))
		self.assertEqual(self.test_case.loggy.POOL_STATE, 0)
		self.assertTrue(hasattr(self.test_case.loggy, "BUFFER"))
		self.assertEqual(self.test_case.loggy.BUFFER, [])








	def test_outer_PoolHoist_name_default_members(self):
		class Outdoor(Transmutable):

			@PoolHoist
			class shmoggy(LibraryContrib):	pass
		class Irma(Outdoor):				pass


		self.test_case = Irma()
		self.assertTrue(hasattr(self.test_case.loggy, "handler"))
		self.assertIsInstance(self.test_case.loggy.handler, StreamHandler)
		self.assertTrue(hasattr(self.test_case.loggy, "init_name"))
		self.assertEqual(self.test_case.loggy.init_name, "fantastic logs and where to contribute them")
		self.assertTrue(hasattr(self.test_case.loggy, "init_level"))
		self.assertEqual(self.test_case.loggy.init_level, 20)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_mode"))
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(hasattr(self.test_case.loggy, "contribfmt"))
		self.assertEqual(

			self.test_case.loggy.contribfmt,
			{
				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s : %(message)s",
				"datefmt": "%d/%m/%Y %H%M",
			}
		)

		self.assertTrue(hasattr(self.test_case.loggy, "forcing_map"))
		self.assertIsInstance(self.test_case.loggy.forcing_map, dict)
		self.assertFalse(self.test_case.loggy.forcing_map)

		self.assertTrue(hasattr(self.test_case.loggy, "handover_map"))
		self.assertIsInstance(self.test_case.loggy.handover_map, dict)
		self.assertEqual(self.test_case.loggy.handover_map, { "Irma": 20 })

		self.assertTrue(hasattr(self.test_case.loggy, "contributor_name"))
		self.assertIsNone(self.test_case.loggy.contributor_name)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_order"))
		self.assertEqual(self.test_case.loggy.handover_order, [ "Irma" ])
		self.assertTrue(hasattr(self.test_case.loggy, "contributor"))
		self.assertIsInstance(self.test_case.loggy.contributor, Logger)

		self.assertTrue(hasattr(self.test_case.loggy, "POOL_TIMER"))
		self.assertEqual(self.test_case.loggy.POOL_TIMER, 5)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_LIMIT"))
		self.assertEqual(self.test_case.loggy.POOL_LIMIT, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_NAME"))
		self.assertIsInstance(self.test_case.loggy.POOL_NAME, str)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_STATE"))
		self.assertEqual(self.test_case.loggy.POOL_STATE, 0)
		self.assertTrue(hasattr(self.test_case.loggy, "BUFFER"))
		self.assertEqual(self.test_case.loggy.BUFFER, [])








	def test_outer_PoolHoist_escalated_default_members(self):

		class Outdoor(Transmutable):
			class Shirma(Transmutable):

				@PoolHoist
				class loggy(LibraryContrib):	pass
		class Irma(Outdoor):					pass


		self.test_case = Irma()
		self.assertTrue(hasattr(self.test_case.loggy, "handler"))
		self.assertIsInstance(self.test_case.loggy.handler, StreamHandler)
		self.assertTrue(hasattr(self.test_case.loggy, "init_name"))
		self.assertEqual(self.test_case.loggy.init_name, "fantastic logs and where to contribute them")
		self.assertTrue(hasattr(self.test_case.loggy, "init_level"))
		self.assertEqual(self.test_case.loggy.init_level, 20)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_mode"))
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(hasattr(self.test_case.loggy, "contribfmt"))
		self.assertEqual(

			self.test_case.loggy.contribfmt,
			{
				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s : %(message)s",
				"datefmt": "%d/%m/%Y %H%M",
			}
		)

		self.assertTrue(hasattr(self.test_case.loggy, "forcing_map"))
		self.assertIsInstance(self.test_case.loggy.forcing_map, dict)
		self.assertFalse(self.test_case.loggy.forcing_map)

		self.assertTrue(hasattr(self.test_case.loggy, "handover_map"))
		self.assertIsInstance(self.test_case.loggy.handover_map, dict)
		self.assertEqual(self.test_case.loggy.handover_map, { "Irma": 20, "Irma.Shirma": 20 })

		self.assertTrue(hasattr(self.test_case.loggy, "contributor_name"))
		self.assertIsNone(self.test_case.loggy.contributor_name)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_order"))
		self.assertEqual(self.test_case.loggy.handover_order, [ "Irma.Shirma", "Irma" ])
		self.assertTrue(hasattr(self.test_case.loggy, "contributor"))
		self.assertIsInstance(self.test_case.loggy.contributor, Logger)

		self.assertTrue(hasattr(self.test_case.loggy, "POOL_TIMER"))
		self.assertEqual(self.test_case.loggy.POOL_TIMER, 5)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_LIMIT"))
		self.assertEqual(self.test_case.loggy.POOL_LIMIT, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_NAME"))
		self.assertIsInstance(self.test_case.loggy.POOL_NAME, str)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_STATE"))
		self.assertEqual(self.test_case.loggy.POOL_STATE, 0)
		self.assertTrue(hasattr(self.test_case.loggy, "BUFFER"))
		self.assertEqual(self.test_case.loggy.BUFFER, [])








	def test_outer_PoolHoist_escalated_name_default_members(self):

		class Outdoor(Transmutable):
			class Shirma(Transmutable):

				@PoolHoist
				class shmoggy(LibraryContrib):	pass
		class Irma(Outdoor):					pass


		self.test_case = Irma()
		self.assertTrue(hasattr(self.test_case.loggy, "handler"))
		self.assertIsInstance(self.test_case.loggy.handler, StreamHandler)
		self.assertTrue(hasattr(self.test_case.loggy, "init_name"))
		self.assertEqual(self.test_case.loggy.init_name, "fantastic logs and where to contribute them")
		self.assertTrue(hasattr(self.test_case.loggy, "init_level"))
		self.assertEqual(self.test_case.loggy.init_level, 20)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_mode"))
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(hasattr(self.test_case.loggy, "contribfmt"))
		self.assertEqual(

			self.test_case.loggy.contribfmt,
			{
				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s : %(message)s",
				"datefmt": "%d/%m/%Y %H%M",
			}
		)

		self.assertTrue(hasattr(self.test_case.loggy, "forcing_map"))
		self.assertIsInstance(self.test_case.loggy.forcing_map, dict)
		self.assertFalse(self.test_case.loggy.forcing_map)

		self.assertTrue(hasattr(self.test_case.loggy, "handover_map"))
		self.assertIsInstance(self.test_case.loggy.handover_map, dict)
		self.assertEqual(self.test_case.loggy.handover_map, { "Irma": 20, "Irma.Shirma": 20 })

		self.assertTrue(hasattr(self.test_case.loggy, "contributor_name"))
		self.assertIsNone(self.test_case.loggy.contributor_name)
		self.assertTrue(hasattr(self.test_case.loggy, "handover_order"))
		self.assertEqual(self.test_case.loggy.handover_order, [ "Irma.Shirma", "Irma" ])
		self.assertTrue(hasattr(self.test_case.loggy, "contributor"))
		self.assertIsInstance(self.test_case.loggy.contributor, Logger)

		self.assertTrue(hasattr(self.test_case.loggy, "POOL_TIMER"))
		self.assertEqual(self.test_case.loggy.POOL_TIMER, 5)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_LIMIT"))
		self.assertEqual(self.test_case.loggy.POOL_LIMIT, 50)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_NAME"))
		self.assertIsInstance(self.test_case.loggy.POOL_NAME, str)
		self.assertTrue(hasattr(self.test_case.loggy, "POOL_STATE"))
		self.assertEqual(self.test_case.loggy.POOL_STATE, 0)
		self.assertTrue(hasattr(self.test_case.loggy, "BUFFER"))
		self.assertEqual(self.test_case.loggy.BUFFER, [])








	def test_debug_augmentor(self):

		debug_aug = str(self.IRMA_ROOT /"debug_aug.loggy")
		self.make_loggy_file(debug_aug)

		class Irma(Transmutable):

			@ContribInterceptingCase.DebugAugmentor
			class loggy(LibraryContrib):

				handler		= debug_aug
				init_name	= "debug_aug"
				init_level	= 10

			class Levels(ContribInterceptingCase.Levels):	pass


		loggy = []
		self.test_case = Irma()
		self.test_case.Levels.debugs()
		self.test_case.Levels.infos()
		self.test_case.Levels.warnings()
		self.test_case.Levels.errors()
		self.test_case.Levels.criticals()


		self.test_case.loggy.close()
		with open(debug_aug) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])


		self.assertEqual(
			loggy,
			[
				"@debug_aug-Irma.Levels.debugs DEBUG : Must be logged (augmented)",
				"@debug_aug-Irma.Levels.debugs INFO : Must be logged",
				"@debug_aug-Irma.Levels.debugs WARNING : Must be logged",
				"@debug_aug-Irma.Levels.debugs ERROR : Must be logged",
				"@debug_aug-Irma.Levels.debugs CRITICAL : Must be logged",

				"@debug_aug-Irma.Levels.infos DEBUG : Must be logged (augmented)",
				"@debug_aug-Irma.Levels.infos INFO : Must be logged",
				"@debug_aug-Irma.Levels.infos WARNING : Must be logged",
				"@debug_aug-Irma.Levels.infos ERROR : Must be logged",
				"@debug_aug-Irma.Levels.infos CRITICAL : Must be logged",

				"@debug_aug-Irma.Levels.warnings DEBUG : Must be logged (augmented)",
				"@debug_aug-Irma.Levels.warnings INFO : Must be logged",
				"@debug_aug-Irma.Levels.warnings WARNING : Must be logged",
				"@debug_aug-Irma.Levels.warnings ERROR : Must be logged",
				"@debug_aug-Irma.Levels.warnings CRITICAL : Must be logged",

				"@debug_aug-Irma.Levels.errors DEBUG : Must be logged (augmented)",
				"@debug_aug-Irma.Levels.errors INFO : Must be logged",
				"@debug_aug-Irma.Levels.errors WARNING : Must be logged",
				"@debug_aug-Irma.Levels.errors ERROR : Must be logged",
				"@debug_aug-Irma.Levels.errors CRITICAL : Must be logged",

				"@debug_aug-Irma.Levels.criticals DEBUG : Must be logged (augmented)",
				"@debug_aug-Irma.Levels.criticals INFO : Must be logged",
				"@debug_aug-Irma.Levels.criticals WARNING : Must be logged",
				"@debug_aug-Irma.Levels.criticals ERROR : Must be logged",
				"@debug_aug-Irma.Levels.criticals CRITICAL : Must be logged",
			]
		)
		if	os.path.isfile(debug_aug): os.remove(debug_aug)




	def test_info_augmentor(self):

		info_aug = str(self.IRMA_ROOT /"info_aug.loggy")
		self.make_loggy_file(info_aug)

		class Irma(Transmutable):

			@ContribInterceptingCase.DebugAugmentor
			@ContribInterceptingCase.InfoAugmentor
			class loggy(LibraryContrib):

				handler		= info_aug
				init_name	= "info_aug"
				init_level	= 10

			class Levels(ContribInterceptingCase.Levels):	pass


		loggy = []
		self.test_case = Irma()
		self.test_case.Levels.debugs()
		self.test_case.Levels.infos()
		self.test_case.Levels.warnings()
		self.test_case.Levels.errors()
		self.test_case.Levels.criticals()


		self.test_case.loggy.close()
		with open(info_aug) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])


		self.assertEqual(
			loggy,
			[
				"@info_aug-Irma.Levels.debugs DEBUG : Must be logged (augmented)",
				"@info_aug-Irma.Levels.debugs INFO : Must be logged (augmented)",
				"@info_aug-Irma.Levels.debugs WARNING : Must be logged",
				"@info_aug-Irma.Levels.debugs ERROR : Must be logged",
				"@info_aug-Irma.Levels.debugs CRITICAL : Must be logged",

				"@info_aug-Irma.Levels.infos DEBUG : Must be logged (augmented)",
				"@info_aug-Irma.Levels.infos INFO : Must be logged (augmented)",
				"@info_aug-Irma.Levels.infos WARNING : Must be logged",
				"@info_aug-Irma.Levels.infos ERROR : Must be logged",
				"@info_aug-Irma.Levels.infos CRITICAL : Must be logged",

				"@info_aug-Irma.Levels.warnings DEBUG : Must be logged (augmented)",
				"@info_aug-Irma.Levels.warnings INFO : Must be logged (augmented)",
				"@info_aug-Irma.Levels.warnings WARNING : Must be logged",
				"@info_aug-Irma.Levels.warnings ERROR : Must be logged",
				"@info_aug-Irma.Levels.warnings CRITICAL : Must be logged",

				"@info_aug-Irma.Levels.errors DEBUG : Must be logged (augmented)",
				"@info_aug-Irma.Levels.errors INFO : Must be logged (augmented)",
				"@info_aug-Irma.Levels.errors WARNING : Must be logged",
				"@info_aug-Irma.Levels.errors ERROR : Must be logged",
				"@info_aug-Irma.Levels.errors CRITICAL : Must be logged",

				"@info_aug-Irma.Levels.criticals DEBUG : Must be logged (augmented)",
				"@info_aug-Irma.Levels.criticals INFO : Must be logged (augmented)",
				"@info_aug-Irma.Levels.criticals WARNING : Must be logged",
				"@info_aug-Irma.Levels.criticals ERROR : Must be logged",
				"@info_aug-Irma.Levels.criticals CRITICAL : Must be logged",
			]
		)
		if	os.path.isfile(info_aug): os.remove(info_aug)




	def test_warn_augmentor(self):

		warn_aug = str(self.IRMA_ROOT /"warn_aug.loggy")
		self.make_loggy_file(warn_aug)

		class Irma(Transmutable):

			@ContribInterceptingCase.WarningAugmentor
			@ContribInterceptingCase.InfoAugmentor
			@ContribInterceptingCase.DebugAugmentor
			@PoolHoist
			class loggy(LibraryContrib):

				handler		= warn_aug
				init_name	= "warn_aug"
				init_level	= 10

			class Levels(ContribInterceptingCase.Levels):	pass


		loggy = []
		self.test_case = Irma()
		self.test_case.Levels.debugs()
		self.test_case.Levels.infos()
		self.test_case.Levels.warnings()
		self.test_case.Levels.errors()
		self.test_case.Levels.criticals()


		self.test_case.loggy.close()
		with open(warn_aug) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])


		self.assertEqual(
			loggy,
			[
				"@warn_aug-Irma.Levels.debugs DEBUG : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.debugs INFO : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.debugs WARNING : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.debugs ERROR : Must be logged",
				"@warn_aug-Irma.Levels.debugs CRITICAL : Must be logged",

				"@warn_aug-Irma.Levels.infos DEBUG : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.infos INFO : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.infos WARNING : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.infos ERROR : Must be logged",
				"@warn_aug-Irma.Levels.infos CRITICAL : Must be logged",

				"@warn_aug-Irma.Levels.warnings DEBUG : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.warnings INFO : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.warnings WARNING : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.warnings ERROR : Must be logged",
				"@warn_aug-Irma.Levels.warnings CRITICAL : Must be logged",

				"@warn_aug-Irma.Levels.errors DEBUG : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.errors INFO : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.errors WARNING : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.errors ERROR : Must be logged",
				"@warn_aug-Irma.Levels.errors CRITICAL : Must be logged",

				"@warn_aug-Irma.Levels.criticals DEBUG : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.criticals INFO : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.criticals WARNING : Must be logged (augmented)",
				"@warn_aug-Irma.Levels.criticals ERROR : Must be logged",
				"@warn_aug-Irma.Levels.criticals CRITICAL : Must be logged",
			]
		)
		if	os.path.isfile(warn_aug): os.remove(warn_aug)




	def test_error_augmentor(self):

		error_aug = str(self.IRMA_ROOT /"error_aug.loggy")
		self.make_loggy_file(error_aug)

		class Irma(Transmutable):

			@ContribInterceptingCase.WarningAugmentor
			@ContribInterceptingCase.InfoAugmentor
			@PoolHoist
			@ContribInterceptingCase.DebugAugmentor
			@ContribInterceptingCase.ErrorAugmentor
			class loggy(LibraryContrib):

				handler		= error_aug
				init_name	= "error_aug"
				init_level	= 10

			class Levels(ContribInterceptingCase.Levels):	pass


		loggy = []
		self.test_case = Irma()
		self.test_case.Levels.debugs()
		self.test_case.Levels.infos()
		self.test_case.Levels.warnings()
		self.test_case.Levels.errors()
		self.test_case.Levels.criticals()


		self.test_case.loggy.close()
		with open(error_aug) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])


		self.assertEqual(
			loggy,
			[
				"@error_aug-Irma.Levels.debugs DEBUG : Must be logged (augmented)",
				"@error_aug-Irma.Levels.debugs INFO : Must be logged (augmented)",
				"@error_aug-Irma.Levels.debugs WARNING : Must be logged (augmented)",
				"@error_aug-Irma.Levels.debugs ERROR : Must be logged (augmented)",
				"@error_aug-Irma.Levels.debugs CRITICAL : Must be logged",

				"@error_aug-Irma.Levels.infos DEBUG : Must be logged (augmented)",
				"@error_aug-Irma.Levels.infos INFO : Must be logged (augmented)",
				"@error_aug-Irma.Levels.infos WARNING : Must be logged (augmented)",
				"@error_aug-Irma.Levels.infos ERROR : Must be logged (augmented)",
				"@error_aug-Irma.Levels.infos CRITICAL : Must be logged",

				"@error_aug-Irma.Levels.warnings DEBUG : Must be logged (augmented)",
				"@error_aug-Irma.Levels.warnings INFO : Must be logged (augmented)",
				"@error_aug-Irma.Levels.warnings WARNING : Must be logged (augmented)",
				"@error_aug-Irma.Levels.warnings ERROR : Must be logged (augmented)",
				"@error_aug-Irma.Levels.warnings CRITICAL : Must be logged",

				"@error_aug-Irma.Levels.errors DEBUG : Must be logged (augmented)",
				"@error_aug-Irma.Levels.errors INFO : Must be logged (augmented)",
				"@error_aug-Irma.Levels.errors WARNING : Must be logged (augmented)",
				"@error_aug-Irma.Levels.errors ERROR : Must be logged (augmented)",
				"@error_aug-Irma.Levels.errors CRITICAL : Must be logged",

				"@error_aug-Irma.Levels.criticals DEBUG : Must be logged (augmented)",
				"@error_aug-Irma.Levels.criticals INFO : Must be logged (augmented)",
				"@error_aug-Irma.Levels.criticals WARNING : Must be logged (augmented)",
				"@error_aug-Irma.Levels.criticals ERROR : Must be logged (augmented)",
				"@error_aug-Irma.Levels.criticals CRITICAL : Must be logged",
			]
		)
		if	os.path.isfile(error_aug): os.remove(error_aug)




	def test_critical_augmentor(self):

		crit_aug = str(self.IRMA_ROOT /"crit_aug.loggy")
		self.make_loggy_file(crit_aug)

		class Irma(Transmutable):

			@PoolHoist
			@ContribInterceptingCase.CriticalAugmentor
			@ContribInterceptingCase.ErrorAugmentor
			@ContribInterceptingCase.WarningAugmentor
			@ContribInterceptingCase.InfoAugmentor
			@ContribInterceptingCase.DebugAugmentor
			class loggy(LibraryContrib):

				handler		= crit_aug
				init_name	= "crit_aug"
				init_level	= 10

			class Levels(ContribInterceptingCase.Levels):	pass


		loggy = []
		self.test_case = Irma()
		self.test_case.Levels.debugs()
		self.test_case.Levels.infos()
		self.test_case.Levels.warnings()
		self.test_case.Levels.errors()
		self.test_case.Levels.criticals()


		self.test_case.loggy.close()
		with open(crit_aug) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])


		self.assertEqual(
			loggy,
			[
				"@crit_aug-Irma.Levels.debugs DEBUG : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.debugs INFO : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.debugs WARNING : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.debugs ERROR : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.debugs CRITICAL : Must be logged (augmented)",

				"@crit_aug-Irma.Levels.infos DEBUG : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.infos INFO : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.infos WARNING : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.infos ERROR : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.infos CRITICAL : Must be logged (augmented)",

				"@crit_aug-Irma.Levels.warnings DEBUG : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.warnings INFO : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.warnings WARNING : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.warnings ERROR : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.warnings CRITICAL : Must be logged (augmented)",

				"@crit_aug-Irma.Levels.errors DEBUG : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.errors INFO : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.errors WARNING : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.errors ERROR : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.errors CRITICAL : Must be logged (augmented)",

				"@crit_aug-Irma.Levels.criticals DEBUG : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.criticals INFO : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.criticals WARNING : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.criticals ERROR : Must be logged (augmented)",
				"@crit_aug-Irma.Levels.criticals CRITICAL : Must be logged (augmented)",
			]
		)
		if	os.path.isfile(crit_aug): os.remove(crit_aug)








	def test_PoolHoist_handover_switch(self):

		pool_handover = str(self.IRMA_ROOT /"pool_handover.loggy")
		self.make_loggy_file(pool_handover)

		class Irma(Transmutable):

			@PoolHoist
			class loggy(LibraryContrib):

				handler			= pool_handover
				init_name		= "pool_handover"
				force_handover	= True

			class Levels(ContribInterceptingCase.Levels):	pass


		self.test_case = Irma()
		self.assertEqual(len(self.test_case.loggy.handover_map),7)
		self.assertEqual(self.test_case.loggy.handover_map["Irma"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.debugs"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.infos"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.warnings"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.errors"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.criticals"],20)
		loggy = []
		self.test_case.Levels.errors()
		self.assertTrue(self.test_case.loggy.handover_mode)
		self.assertFalse(self.test_case.loggy.handover_off)

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.errors INFO : Must be logged",
				"@pool_handover-Irma.Levels.errors WARNING : Must be logged",
				"@pool_handover-Irma.Levels.errors ERROR : Must be logged",
				"@pool_handover-Irma.Levels.errors CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.loggy.handover_mode = False
		self.test_case.loggy.force("*errors", level=10)
		self.assertEqual(len(self.test_case.loggy.handover_map),7)
		self.assertEqual(self.test_case.loggy.handover_map["Irma"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.debugs"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.infos"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.warnings"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.errors"],10)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.criticals"],20)
		self.test_case.Levels.errors()
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(self.test_case.loggy.handover_off)

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.errors DEBUG : Must be logged",
				"@pool_handover-Irma.Levels.errors INFO : Must be logged",
				"@pool_handover-Irma.Levels.errors WARNING : Must be logged",
				"@pool_handover-Irma.Levels.errors ERROR : Must be logged",
				"@pool_handover-Irma.Levels.errors CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.loggy.handover_mode = True
		self.test_case.Levels.errors()
		self.assertTrue(self.test_case.loggy.handover_mode)
		self.assertFalse(self.test_case.loggy.handover_off)

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.errors DEBUG : Must be logged",
				"@pool_handover-Irma.Levels.errors INFO : Must be logged",
				"@pool_handover-Irma.Levels.errors WARNING : Must be logged",
				"@pool_handover-Irma.Levels.errors ERROR : Must be logged",
				"@pool_handover-Irma.Levels.errors CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.loggy.force("*warnings", level=10)
		self.assertEqual(len(self.test_case.loggy.handover_map),7)
		self.assertEqual(self.test_case.loggy.handover_map["Irma"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.debugs"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.infos"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.warnings"],10)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.errors"],10)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.criticals"],20)
		self.test_case.Levels.warnings()

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.warnings DEBUG : Must be logged",
				"@pool_handover-Irma.Levels.warnings INFO : Must be logged",
				"@pool_handover-Irma.Levels.warnings WARNING : Must be logged",
				"@pool_handover-Irma.Levels.warnings ERROR : Must be logged",
				"@pool_handover-Irma.Levels.warnings CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.Levels.criticals()

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.criticals INFO : Must be logged",
				"@pool_handover-Irma.Levels.criticals WARNING : Must be logged",
				"@pool_handover-Irma.Levels.criticals ERROR : Must be logged",
				"@pool_handover-Irma.Levels.criticals CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.Levels.warnings()

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.warnings DEBUG : Must be logged",
				"@pool_handover-Irma.Levels.warnings INFO : Must be logged",
				"@pool_handover-Irma.Levels.warnings WARNING : Must be logged",
				"@pool_handover-Irma.Levels.warnings ERROR : Must be logged",
				"@pool_handover-Irma.Levels.warnings CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.loggy.handover_mode = False
		self.test_case.loggy.force("*warnings", level=20)
		self.assertEqual(len(self.test_case.loggy.handover_map),7)
		self.assertEqual(self.test_case.loggy.handover_map["Irma"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.debugs"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.infos"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.warnings"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.errors"],10)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.criticals"],20)
		self.test_case.Levels.warnings()
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(self.test_case.loggy.handover_off)

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover INFO : Must be logged",
				"@pool_handover WARNING : Must be logged",
				"@pool_handover ERROR : Must be logged",
				"@pool_handover CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.loggy.handover_mode = True
		self.test_case.Levels.errors()
		self.assertTrue(self.test_case.loggy.handover_mode)
		self.assertFalse(self.test_case.loggy.handover_off)

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.errors DEBUG : Must be logged",
				"@pool_handover-Irma.Levels.errors INFO : Must be logged",
				"@pool_handover-Irma.Levels.errors WARNING : Must be logged",
				"@pool_handover-Irma.Levels.errors ERROR : Must be logged",
				"@pool_handover-Irma.Levels.errors CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.loggy.handover_mode = False
		self.test_case.Levels.warnings()
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(self.test_case.loggy.handover_off)

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover INFO : Must be logged",
				"@pool_handover WARNING : Must be logged",
				"@pool_handover ERROR : Must be logged",
				"@pool_handover CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.loggy.handover_mode = True
		self.test_case.loggy.force("*warnings", level=10)
		self.assertEqual(len(self.test_case.loggy.handover_map),7)
		self.assertEqual(self.test_case.loggy.handover_map["Irma"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.debugs"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.infos"],20)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.warnings"],10)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.errors"],10)
		self.assertEqual(self.test_case.loggy.handover_map["Irma.Levels.criticals"],20)
		self.test_case.Levels.warnings()
		self.assertTrue(self.test_case.loggy.handover_mode)
		self.assertFalse(self.test_case.loggy.handover_off)

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.warnings DEBUG : Must be logged",
				"@pool_handover-Irma.Levels.warnings INFO : Must be logged",
				"@pool_handover-Irma.Levels.warnings WARNING : Must be logged",
				"@pool_handover-Irma.Levels.warnings ERROR : Must be logged",
				"@pool_handover-Irma.Levels.warnings CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.loggy.handover_mode = False
		self.test_case.Levels.warnings()
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(self.test_case.loggy.handover_off)

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.warnings DEBUG : Must be logged",
				"@pool_handover-Irma.Levels.warnings INFO : Must be logged",
				"@pool_handover-Irma.Levels.warnings WARNING : Must be logged",
				"@pool_handover-Irma.Levels.warnings ERROR : Must be logged",
				"@pool_handover-Irma.Levels.warnings CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.loggy.handover_mode = True
		self.test_case.Levels.errors()
		self.assertTrue(self.test_case.loggy.handover_mode)
		self.assertFalse(self.test_case.loggy.handover_off)

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.errors DEBUG : Must be logged",
				"@pool_handover-Irma.Levels.errors INFO : Must be logged",
				"@pool_handover-Irma.Levels.errors WARNING : Must be logged",
				"@pool_handover-Irma.Levels.errors ERROR : Must be logged",
				"@pool_handover-Irma.Levels.errors CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.loggy.handover_mode = False
		self.test_case.Levels.warnings()
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(self.test_case.loggy.handover_off)

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.warnings DEBUG : Must be logged",
				"@pool_handover-Irma.Levels.warnings INFO : Must be logged",
				"@pool_handover-Irma.Levels.warnings WARNING : Must be logged",
				"@pool_handover-Irma.Levels.warnings ERROR : Must be logged",
				"@pool_handover-Irma.Levels.warnings CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.loggy.handover_mode = True
		self.test_case.Levels.infos()
		self.assertTrue(self.test_case.loggy.handover_mode)
		self.assertFalse(self.test_case.loggy.handover_off)

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.infos INFO : Must be logged",
				"@pool_handover-Irma.Levels.infos WARNING : Must be logged",
				"@pool_handover-Irma.Levels.infos ERROR : Must be logged",
				"@pool_handover-Irma.Levels.infos CRITICAL : Must be logged",
			]
		)


		loggy = []
		self.make_loggy_file(pool_handover)
		self.test_case.loggy.handover_mode = False
		self.test_case.Levels.warnings()
		self.assertFalse(self.test_case.loggy.handover_mode)
		self.assertTrue(self.test_case.loggy.handover_off)

		self.test_case.loggy.close()
		with open(pool_handover) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(
			loggy,
			[
				"@pool_handover-Irma.Levels.warnings DEBUG : Must be logged",
				"@pool_handover-Irma.Levels.warnings INFO : Must be logged",
				"@pool_handover-Irma.Levels.warnings WARNING : Must be logged",
				"@pool_handover-Irma.Levels.warnings ERROR : Must be logged",
				"@pool_handover-Irma.Levels.warnings CRITICAL : Must be logged",
			]
		)
		if	os.path.isfile(pool_handover): os.remove(pool_handover)








	def test_info_hoist_to_writer1(self):

		ih2w1 = str(self.IRMA_ROOT /"ih2w1.loggy")
		self.make_loggy_file(ih2w1)


		ih2w1_release = str(self.IRMA_ROOT /"ih2w1.release")
		if os.path.isfile(ih2w1_release): os.remove(ih2w1_release)
		self.assertFalse(os.path.isfile(ih2w1_release))


		class Irma(Transmutable):
			def __init__(self, *args, **kwargs):

				super().__init__(*args, **kwargs)
				self.loggy.info("Irma entered the library")

			@ContribInterceptingCase.InfoHoist
			@ContribInterceptingCase.ReleaseWriter1
			@PoolHoist
			class loggy(LibraryContrib):

				handler		= ih2w1
				init_name	= "ih2w1"
				init_level	= 10

			class Levels(ContribInterceptingCase.Levels):	pass
			def __call__(self):

				self.Levels.debugs()
				self.Levels.infos()
				self.Levels.warnings()
				self.Levels.errors()
				self.Levels.criticals()


		loggy = []
		self.test_case = Irma()
		self.test_case()


		self.assertTrue(self.test_case.loggy.POOL_STATE)
		sleep(10.5)
		self.assertFalse(self.test_case.loggy.POOL_STATE)


		self.test_case.loggy.close()
		with open(ih2w1) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])


		self.assertEqual(
			loggy,
			[
				"@ih2w1-Irma INFO : Irma entered the library",

				"@ih2w1-loggy.pool DEBUG : (pool-id-0) 24 symbols message inserted",
				"@ih2w1-loggy.pool DEBUG : (pool-id-0) buffer extended to 1 item",
				"@ih2w1-loggy.pool DEBUG : (pool-id-0) starting new thread",


				"@ih2w1-Irma.Levels.debugs DEBUG : Must be logged",
				"@ih2w1-Irma.Levels.debugs INFO : Must be logged",

				"@ih2w1-loggy.pool DEBUG : (pool-id-1) 14 symbols message inserted",
				"@ih2w1-loggy.pool DEBUG : (pool-id-1) buffer extended to 2 items",

				"@ih2w1-Irma.Levels.debugs WARNING : Must be logged",
				"@ih2w1-Irma.Levels.debugs ERROR : Must be logged",
				"@ih2w1-Irma.Levels.debugs CRITICAL : Must be logged",

				"@ih2w1-Irma.Levels.infos DEBUG : Must be logged",
				"@ih2w1-Irma.Levels.infos INFO : Must be logged",
				"@ih2w1-Irma.Levels.infos WARNING : Must be logged",
				"@ih2w1-Irma.Levels.infos ERROR : Must be logged",
				"@ih2w1-Irma.Levels.infos CRITICAL : Must be logged",
				"@ih2w1-Irma.Levels.warnings DEBUG : Must be logged",
				"@ih2w1-Irma.Levels.warnings INFO : Must be logged",
				"@ih2w1-Irma.Levels.warnings WARNING : Must be logged",
				"@ih2w1-Irma.Levels.warnings ERROR : Must be logged",
				"@ih2w1-Irma.Levels.warnings CRITICAL : Must be logged",
				"@ih2w1-Irma.Levels.errors DEBUG : Must be logged",
				"@ih2w1-Irma.Levels.errors INFO : Must be logged",
				"@ih2w1-Irma.Levels.errors WARNING : Must be logged",
				"@ih2w1-Irma.Levels.errors ERROR : Must be logged",
				"@ih2w1-Irma.Levels.errors CRITICAL : Must be logged",
				"@ih2w1-Irma.Levels.criticals DEBUG : Must be logged",
				"@ih2w1-Irma.Levels.criticals INFO : Must be logged",
				"@ih2w1-Irma.Levels.criticals WARNING : Must be logged",
				"@ih2w1-Irma.Levels.criticals ERROR : Must be logged",
				"@ih2w1-Irma.Levels.criticals CRITICAL : Must be logged",

				"@ih2w1-loggy.pool DEBUG : (pool-id-0) timer extended",
				"@ih2w1-loggy.pool DEBUG : (pool-id-0) was 10 seconds",
				"@ih2w1-loggy.pool DEBUG : (pool-id-0) was 2 items",
			]
		)


		self.assertTrue(os.path.isfile(ih2w1_release))
		buffer_content = []


		self.test_case.loggy.close()
		with open(ih2w1_release) as buffer:
			for line in buffer : buffer_content.append(line)


		self.assertEqual(buffer_content,[ "Irma entered the library\n", "Must be logged\n" ])
		if	os.path.isfile(ih2w1):			os.remove(ih2w1)
		if	os.path.isfile(ih2w1_release):	os.remove(ih2w1_release)








	@unittest.skipIf(os.name == "nt", "cannot test termination, cause windows cannot fork")
	def test_info_forked_hoist_to_writer2(self):

		ih2w2 = str(self.IRMA_ROOT /"ih2w2.loggy")
		self.make_loggy_file(ih2w2)


		ih2w2_release = str(self.IRMA_ROOT /"ih2w2.release")
		if os.path.isfile(ih2w2_release): os.remove(ih2w2_release)
		self.assertFalse(os.path.isfile(ih2w2_release))


		@DIRTtimer(T=1.5)
		class Irma(Transmutable):
			def __init__(self, *args, **kwargs):

				super().__init__(*args, **kwargs)
				self.loggy.info("Irma entered the library")

			@ContribInterceptingCase.ReleaseWriter2
			@ContribInterceptingCase.InfoHoist
			class loggy(LibraryContrib):

				handler		= ih2w2
				init_name	= "ih2w2"
				init_level	= 10
				pool_timer	= .5

			class Levels(ContribInterceptingCase.Levels):	pass
			def __call__(self):

				self.Levels.debugs()
				self.loggy.info("OOH EEH OOH AH AH")
				self.Levels.infos()
				self.Levels.warnings()
				self.loggy.info("TING TANG")
				self.Levels.errors()
				self.Levels.criticals()
				self.loggy.info("WALA WALA BING BANG")


		loggy = []
		self.test_case = Irma()
		self.test_case()
		sleep(1.5)


		self.test_case.loggy.close()
		with open(ih2w2) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])


		self.assertEqual(
			loggy,
			[
				"@ih2w2-Irma INFO : Irma entered the library",

				"@ih2w2-loggy.pool DEBUG : (pool-id-0) 24 symbols message inserted",
				"@ih2w2-loggy.pool DEBUG : (pool-id-0) buffer extended to 1 item",
				"@ih2w2-loggy.pool DEBUG : (pool-id-0) starting new thread",

				"@ih2w2-Irma DEBUG : Delay start timer 0.0 seconds",
				"@ih2w2-Irma DEBUG : Interval delay 0.0 seconds",
				"@ih2w2-Irma DEBUG : Repetition counter 1 time",
				"@ih2w2-Irma DEBUG : Termination timer 1.5 seconds",
				"@ih2w2-Irma DEBUG : Caller arguments: ()",
				r"@ih2w2-Irma DEBUG : Caller keyword arguments: {}",
				"@ih2w2-Irma DEBUG : DIRT iteration 1",
				"@ih2w2-Irma INFO : Starting 1.5 seconds timer for Irma",

				"@ih2w2-loggy.pool DEBUG : (pool-id-1) 35 symbols message inserted",
				"@ih2w2-loggy.pool DEBUG : (pool-id-1) buffer extended to 2 items",

				"@ih2w2-Irma.Levels.debugs DEBUG : Must be logged",
				"@ih2w2-Irma.Levels.debugs INFO : Must be logged",

				"@ih2w2-loggy.pool DEBUG : (pool-id-2) switching from MainProcess to Process-1",
				"@ih2w2-loggy.pool DEBUG : (pool-id-2) wiping 2 items buffer",
				"@ih2w2-loggy.pool DEBUG : (pool-id-2) 14 symbols message inserted",
				"@ih2w2-loggy.pool DEBUG : (pool-id-2) buffer extended to 1 item",
				"@ih2w2-loggy.pool DEBUG : (pool-id-2) starting new thread",

				"@ih2w2-Irma.Levels.debugs WARNING : Must be logged",
				"@ih2w2-Irma.Levels.debugs ERROR : Must be logged",
				"@ih2w2-Irma.Levels.debugs CRITICAL : Must be logged",

				"@ih2w2-Irma INFO : OOH EEH OOH AH AH",

				"@ih2w2-loggy.pool DEBUG : (pool-id-3) 17 symbols message inserted",
				"@ih2w2-loggy.pool DEBUG : (pool-id-3) buffer extended to 2 items",

				"@ih2w2-Irma.Levels.infos DEBUG : Must be logged",
				"@ih2w2-Irma.Levels.infos INFO : Must be logged",
				"@ih2w2-Irma.Levels.infos WARNING : Must be logged",
				"@ih2w2-Irma.Levels.infos ERROR : Must be logged",
				"@ih2w2-Irma.Levels.infos CRITICAL : Must be logged",
				"@ih2w2-Irma.Levels.warnings DEBUG : Must be logged",
				"@ih2w2-Irma.Levels.warnings INFO : Must be logged",
				"@ih2w2-Irma.Levels.warnings WARNING : Must be logged",
				"@ih2w2-Irma.Levels.warnings ERROR : Must be logged",
				"@ih2w2-Irma.Levels.warnings CRITICAL : Must be logged",

				"@ih2w2-Irma INFO : TING TANG",

				"@ih2w2-loggy.pool DEBUG : (pool-id-4) 9 symbols message inserted",
				"@ih2w2-loggy.pool DEBUG : (pool-id-4) buffer extended to 3 items",

				"@ih2w2-Irma.Levels.errors DEBUG : Must be logged",
				"@ih2w2-Irma.Levels.errors INFO : Must be logged",
				"@ih2w2-Irma.Levels.errors WARNING : Must be logged",
				"@ih2w2-Irma.Levels.errors ERROR : Must be logged",
				"@ih2w2-Irma.Levels.errors CRITICAL : Must be logged",
				"@ih2w2-Irma.Levels.criticals DEBUG : Must be logged",
				"@ih2w2-Irma.Levels.criticals INFO : Must be logged",
				"@ih2w2-Irma.Levels.criticals WARNING : Must be logged",
				"@ih2w2-Irma.Levels.criticals ERROR : Must be logged",
				"@ih2w2-Irma.Levels.criticals CRITICAL : Must be logged",

				"@ih2w2-Irma INFO : WALA WALA BING BANG",

				"@ih2w2-loggy.pool DEBUG : (pool-id-5) 19 symbols message inserted",
				"@ih2w2-loggy.pool DEBUG : (pool-id-5) buffer extended to 4 items",

				"@ih2w2-loggy.pool DEBUG : (pool-id-0) timer extended",
				"@ih2w2-loggy.pool DEBUG : (pool-id-0) was 1.0 seconds",
				"@ih2w2-loggy.pool DEBUG : (pool-id-0) was 2 items",

				"@ih2w2-loggy.pool DEBUG : (pool-id-2) timer extended",
				"@ih2w2-loggy.pool DEBUG : (pool-id-2) was 1.0 seconds",
				"@ih2w2-loggy.pool DEBUG : (pool-id-2) was 4 items",

				"@ih2w2-Irma DEBUG : Process Irma terminated",
			]
		)


		self.assertTrue(os.path.isfile(ih2w2_release))
		buffer_content = []


		self.test_case.loggy.close()
		with open(ih2w2_release) as buffer:
			for line in buffer : buffer_content.append(line)


		self.assertEqual(
			buffer_content,
			[
				"Must be logged\n",
				"OOH EEH OOH AH AH\n",
				"TING TANG\n",
				"WALA WALA BING BANG\n",
			]
		)
		if	os.path.isfile(ih2w2):			os.remove(ih2w2)
		if	os.path.isfile(ih2w2_release):	os.remove(ih2w2_release)








	def test_pool_limit(self):
		class Irma(Transmutable):

			@PoolHoist
			@ContribInterceptingCase.InfoHoist
			class loggy(LibraryContrib):

				handler		= self.INTERCEPT_HANDLER
				init_name	= "pool_limit"
				init_level	= 10
				pool_timer	= .1
				pool_limit	= 1

			class Levels(ContribInterceptingCase.Levels):	pass

		with self.assertLogs("pool_limit", 10) as case_loggy:

			self.test_case = Irma()
			for i in range(20):

				self.test_case.loggy.info(f"Current is {i}")
				sleep(.08)

		sleep(1)
		self.assertIn("DEBUG:pool_limit:(pool-id-0) 12 symbols message inserted", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-0) buffer extended to 1 item", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-0) starting new thread", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-1) 12 symbols message inserted", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-1) buffer extended to 2 items", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-2) 12 symbols message inserted", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-2) buffer extended to 3 items", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-3) 12 symbols message inserted", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-3) buffer extended to 4 items", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-4) 12 symbols message inserted", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-4) buffer extended to 5 items", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-5) 12 symbols message inserted", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-5) buffer extended to 6 items", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-6) 12 symbols message inserted", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-6) buffer extended to 7 items", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-7) 12 symbols message inserted", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-7) buffer extended to 8 items", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-8) 12 symbols message inserted", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-8) buffer extended to 9 items", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-9) 12 symbols message inserted", case_loggy.output)
		self.assertIn("DEBUG:pool_limit:(pool-id-9) buffer extended to 10 items", case_loggy.output)

		# Following spaghetti is the monkey-patching for floats to be calculated very funny.
		# Originally was found that current test "loggy.pool" "buffer_time" will not be able to
		# succesfully sum up to 1.0, so for some different cases below handlings.
		try:			self.assertIn("DEBUG:pool_limit:(pool-id-0) was 10 items", case_loggy.output)
		except:
			try:		self.assertIn("DEBUG:pool_limit:(pool-id-0) was 11 items", case_loggy.output)
			except:
				try:	self.assertIn("DEBUG:pool_limit:(pool-id-0) was 12 items", case_loggy.output)
				except:	self.assertIn("DEBUG:pool_limit:(pool-id-0) was 13 items", case_loggy.output)








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	time								import sleep
from	shutil								import rmtree
from	pygwarts.tests.hagrid				import MediumSet
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.hagrid.thrivables			import Tree
from	pygwarts.hagrid.thrivables			import Copse
from	pygwarts.hagrid.sprouts				import fssprout
from	pygwarts.hagrid.planting			import Flourish
from	pygwarts.hagrid.planting.twigs		import TwigThrive
from	pygwarts.hagrid.planting.leafs		import LeafGrowth
from	pygwarts.hagrid.planting.peels		import GrowingPeel
from	pygwarts.hagrid.planting.peeks		import DraftPeek
from	pygwarts.hagrid.bloom.twigs			import Germination
from	pygwarts.hagrid.bloom.leafs			import Rejuvenation
from	pygwarts.hagrid.cultivation.sifting	import SiftingController








class IntergrowthCase(MediumSet):

	"""
		Intergrowth scheme when there are two flourish with vice versa sprout and bough
	"""

	@classmethod
	def tearDownClass(cls):
		cls.clean(cls)

		if	cls.clean_up:
			if	os.path.isfile(cls.MEDIUM_HANDLER_9): os.remove(cls.MEDIUM_HANDLER_9)

	@classmethod
	def setUpClass(cls):

		super().setUpClass()
		cls.make_loggy_file(cls, cls.MEDIUM_HANDLER_9)
		cls.mss1_flower = os.path.join(cls.MEDIUM_SET_SPROUT_1, os.path.basename(cls.f_msb1_flower))


	def test_separated_walks(self):
		class Sakura(Tree):

			bough	= self.MEDIUM_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.MEDIUM_HANDLER_9
				init_name	= "Intergrowth-1"

			@GrowingPeel
			class thrive(TwigThrive):	pass
			class folders(Germination):	pass

			@GrowingPeel
			@DraftPeek(renew=False)
			class grow(LeafGrowth):		pass
			class files(Rejuvenation):	pass

			@fssprout(self.MEDIUM_SET_SPROUT_1)
			class walk(Flourish):		pass

		class Sequoia(Tree):

			bough	= self.MEDIUM_SET_SPROUT_1
			class loggy(LibraryContrib):

				handler		= self.MEDIUM_HANDLER_9
				init_name	= "Intergrowth-2"

			@GrowingPeel
			class thrive(TwigThrive):	pass
			class folders(Germination):	pass

			@GrowingPeel
			@DraftPeek(renew=False)
			class grow(LeafGrowth):		pass
			class files(Rejuvenation):	pass

			@fssprout(self.MEDIUM_SET_BOUGH_1)
			class walk(Flourish):		pass


		if	not os.path.isfile(self.f_mss1_md_rock):			self.fmake(self.f_mss1_md_rock)
		if	not os.path.isfile(self.f_mss1_md_paper):			self.fmake(self.f_mss1_md_paper)
		if	not os.path.isfile(self.f_mss1_md_scissors):		self.fmake(self.f_mss1_md_scissors)
		if	not os.path.isfile(self.f_mss1_md_gc_sadmovies):	self.fmake(self.f_mss1_md_gc_sadmovies)
		if	not os.path.isfile(self.f_mss1_md_gc_dumbasses):	self.fmake(self.f_mss1_md_gc_dumbasses)
		if	not os.path.isfile(self.f_mss1_md_gc_dumbasses):	self.fmake(self.f_mss1_md_gc_dumbasses)
		if	os.path.isdir(self.tg1_d_mss1_md_gc):				rmtree(self.tg1_d_mss1_md_gc)
		if	os.path.isdir(self.d_mss1_c_good):					rmtree(self.d_mss1_c_good)
		if	os.path.isdir(self.d_mss1_c_notgood):				rmtree(self.d_mss1_c_notgood)
		if	not os.path.isdir(self.d_mss1_c_bad):				os.makedirs(self.d_mss1_c_bad)
		if	not os.path.isdir(self.d_mss1_c_notbad):			os.makedirs(self.d_mss1_c_notbad)
		if	not os.path.isdir(self.d_mss1_c_notnot):			os.makedirs(self.d_mss1_c_notnot)
		if	not os.path.isdir(self.d_mss1_c_badbad):			os.makedirs(self.d_mss1_c_badbad)
		if	not os.path.isdir(self.d_mss1_c_badgood):			os.makedirs(self.d_mss1_c_badgood)
		if	not os.path.isdir(self.d_mss1_c_goodbad):			os.makedirs(self.d_mss1_c_goodbad)

		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertFalse(os.path.isdir(self.d_mss1_c_good))
		self.assertFalse(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))


		if	os.path.isfile(self.tg1_f_mss1_md_rock):			os.remove(self.tg1_f_mss1_md_rock)
		if	os.path.isfile(self.tg1_f_mss1_md_paper):			os.remove(self.tg1_f_mss1_md_paper)
		if	os.path.isfile(self.tg1_f_mss1_md_gc_sadmovies):	os.remove(self.tg1_f_mss1_md_gc_sadmovies)
		if	os.path.isfile(self.tg1_f_mss1_md_gc_dumbasses):	os.remove(self.tg1_f_mss1_md_gc_dumbasses)
		if	os.path.isfile(self.tg1_f_mss1_md_gc_onion):		os.remove(self.tg1_f_mss1_md_gc_onion)
		if	not os.path.isdir(self.tg1_d_mss1_c_good):			os.makedirs(self.tg1_d_mss1_c_good)
		if	not os.path.isdir(self.tg1_d_mss1_c_notgood):		os.makedirs(self.tg1_d_mss1_c_notgood)
		if	not os.path.isfile(self.tg1_f_mss1_md_scissors):	self.fmake(self.tg1_f_mss1_md_scissors)
		if	os.path.isdir(self.tg1_d_mss1_c_bad):				rmtree(self.tg1_d_mss1_c_bad)
		if	os.path.isdir(self.tg1_d_mss1_c_notbad):			rmtree(self.tg1_d_mss1_c_notbad)
		if	os.path.isdir(self.tg1_d_mss1_c_notnot):			rmtree(self.tg1_d_mss1_c_notnot)
		if	os.path.isdir(self.tg1_d_mss1_c_badbad):			rmtree(self.tg1_d_mss1_c_badbad)
		if	os.path.isdir(self.tg1_d_mss1_c_badgood):			rmtree(self.tg1_d_mss1_c_badgood)
		if	os.path.isdir(self.tg1_d_mss1_c_goodbad):			rmtree(self.tg1_d_mss1_c_goodbad)

		# First sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_md))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_rock))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_scissors))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_md_gc))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_gc_sadmovies))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_gc_dumbasses))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_bad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notbad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notnot))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badbad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_goodbad))
		self.assertTrue(os.path.isfile(self.f_msb1_flower))


		with self.assertLogs("Intergrowth-1", 20) as case_loggy:

			self.test_case = Sakura()
			self.test_case.walk()

		self.assertCountEqual(
			case_loggy.output,
			[
				f"INFO:Intergrowth-1:Thrived twig \"{self.tg1_d_mss1_md_gc}\"",
				f"INFO:Intergrowth-1:Grown leaf \"{self.tg1_f_mss1_md_rock}\"",
				f"INFO:Intergrowth-1:Grown leaf \"{self.tg1_f_mss1_md_paper}\"",
				f"INFO:Intergrowth-1:Thrived twig \"{self.tg1_d_mss1_c_notnot}\"",
				f"INFO:Intergrowth-1:Thrived twig \"{self.tg1_d_mss1_c_bad}\"",
				f"INFO:Intergrowth-1:Thrived twig \"{self.tg1_d_mss1_c_notbad}\"",
				f"INFO:Intergrowth-1:Thrived twig \"{self.tg1_d_mss1_c_badbad}\"",
				f"INFO:Intergrowth-1:Thrived twig \"{self.tg1_d_mss1_c_badgood}\"",
				f"INFO:Intergrowth-1:Thrived twig \"{self.tg1_d_mss1_c_goodbad}\"",
				f"INFO:Intergrowth-1:Grown leaf \"{self.tg1_f_mss1_md_gc_dumbasses}\"",
				f"INFO:Intergrowth-1:Grown leaf \"{self.tg1_f_mss1_md_gc_sadmovies}\"",
				f"INFO:Intergrowth-1:Grown leaf \"{self.tg1_f_mss1_md_gc_onion}\"",
			]
		)


		with self.assertLogs("Intergrowth-2", 20) as case_loggy:

			self.test_case.loggy.close()
			self.test_case = Sequoia()
			self.test_case.walk()

		self.assertCountEqual(
			case_loggy.output,
			[
				f"INFO:Intergrowth-2:Grown leaf \"{self.mss1_flower}\"",
				f"INFO:Intergrowth-2:Thrived twig \"{self.d_mss1_c_good}\"",
				f"INFO:Intergrowth-2:Thrived twig \"{self.d_mss1_c_notgood}\"",
			]
		)


		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))


		# First sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_md))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_scissors))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_md_gc))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_goodbad))
		self.assertTrue(os.path.isfile(self.f_msb1_flower))








	def test_single_walk(self):

		class Forest(Copse):
			class loggy(LibraryContrib):

				handler		= self.MEDIUM_HANDLER_9
				init_name	= "Intergrowth"

			class Sakura(Tree):

				bough = self.MEDIUM_SET_BOUGH_1
				class leafs(SiftingController):

					include = (

						( rf"{self.MEDIUM_SET_SPROUT_1}/.+", )
						if os.name == "posix" else
						( self.MEDIUM_SET_SPROUT_1.replace("\\", "\\\\") + r"\\.+", )
					)
				class twigs(SiftingController):

					include = (

						( rf"{self.MEDIUM_SET_SPROUT_1}/.+", )
						if os.name == "posix" else
						( self.MEDIUM_SET_SPROUT_1.replace("\\", "\\\\") + r"\\.+", )
					)

			class Sequoia(Tree):

				bough = self.MEDIUM_SET_SPROUT_1
				class leafs(SiftingController):

					include = (

						( rf"{self.MEDIUM_SET_BOUGH_1}/.+", )
						if os.name == "posix" else
						( self.MEDIUM_SET_BOUGH_1.replace("\\", "\\\\") + r"\\.+", )
					)
				class twigs(SiftingController):

					include = (

						( rf"{self.MEDIUM_SET_BOUGH_1}/.+", )
						if os.name == "posix" else
						( self.MEDIUM_SET_BOUGH_1.replace("\\", "\\\\") + r"\\.+", )
					)

			@GrowingPeel
			class thrive(TwigThrive):	pass
			class folders(Germination):	pass

			# DraftPeek comparator mitigates the effect when rejuved folder becomes a sprout and
			# files seams to be newer then thier last sources, so one hour (for example) gap
			# must ensure there'll be no rewriting
			@GrowingPeel
			@DraftPeek(renew=False, comparator=(lambda F,S : 360 <(F -S)))
			class grow(LeafGrowth):		pass
			class files(Rejuvenation):	pass

			# Basically it might be two separated Flourish objects in one Copse or for different Trees.
			# Basically.
			@fssprout(self.MEDIUM_SET_BOUGH_1)
			@fssprout(self.MEDIUM_SET_SPROUT_1)
			class walk(Flourish):		pass


		self.clean()


		if	not os.path.isfile(self.f_mss1_md_rock):			self.fmake(self.f_mss1_md_rock)
		if	not os.path.isfile(self.f_mss1_md_paper):			self.fmake(self.f_mss1_md_paper)
		if	not os.path.isfile(self.f_mss1_md_scissors):		self.fmake(self.f_mss1_md_scissors)
		if	not os.path.isfile(self.f_mss1_md_gc_sadmovies):	self.fmake(self.f_mss1_md_gc_sadmovies)
		if	not os.path.isfile(self.f_mss1_md_gc_dumbasses):	self.fmake(self.f_mss1_md_gc_dumbasses)
		if	not os.path.isfile(self.f_mss1_md_gc_onion):		self.fmake(self.f_mss1_md_gc_onion)
		if	os.path.isdir(self.tg1_d_mss1_md_gc):				rmtree(self.tg1_d_mss1_md_gc)
		if	os.path.isdir(self.d_mss1_c_good):					rmtree(self.d_mss1_c_good)
		if	os.path.isdir(self.d_mss1_c_notgood):				rmtree(self.d_mss1_c_notgood)
		if	not os.path.isdir(self.d_mss1_c_bad):				os.makedirs(self.d_mss1_c_bad)
		if	not os.path.isdir(self.d_mss1_c_notbad):			os.makedirs(self.d_mss1_c_notbad)
		if	not os.path.isdir(self.d_mss1_c_notnot):			os.makedirs(self.d_mss1_c_notnot)
		if	not os.path.isdir(self.d_mss1_c_badbad):			os.makedirs(self.d_mss1_c_badbad)
		if	not os.path.isdir(self.d_mss1_c_badgood):			os.makedirs(self.d_mss1_c_badgood)
		if	not os.path.isdir(self.d_mss1_c_goodbad):			os.makedirs(self.d_mss1_c_goodbad)

		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertFalse(os.path.isdir(self.d_mss1_c_good))
		self.assertFalse(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))


		sleep(1.1)


		if	os.path.isfile(self.tg1_f_mss1_md_rock):			os.remove(self.tg1_f_mss1_md_rock)
		if	os.path.isfile(self.tg1_f_mss1_md_paper):			os.remove(self.tg1_f_mss1_md_paper)
		if	os.path.isfile(self.tg1_f_mss1_md_gc_sadmovies):	os.remove(self.tg1_f_mss1_md_gc_sadmovies)
		if	os.path.isfile(self.tg1_f_mss1_md_gc_dumbasses):	os.remove(self.tg1_f_mss1_md_gc_dumbasses)
		if	os.path.isfile(self.tg1_f_mss1_md_gc_onion):		os.remove(self.tg1_f_mss1_md_gc_onion)
		if	not os.path.isdir(self.tg1_d_mss1_c_good):			os.makedirs(self.tg1_d_mss1_c_good)
		if	not os.path.isdir(self.tg1_d_mss1_c_notgood):		os.makedirs(self.tg1_d_mss1_c_notgood)
		if	not os.path.isfile(self.tg1_f_mss1_md_scissors):	self.fmake(self.tg1_f_mss1_md_scissors)
		if	os.path.isdir(self.tg1_d_mss1_c_bad):				rmtree(self.tg1_d_mss1_c_bad)
		if	os.path.isdir(self.tg1_d_mss1_c_notbad):			rmtree(self.tg1_d_mss1_c_notbad)
		if	os.path.isdir(self.tg1_d_mss1_c_notnot):			rmtree(self.tg1_d_mss1_c_notnot)
		if	os.path.isdir(self.tg1_d_mss1_c_badbad):			rmtree(self.tg1_d_mss1_c_badbad)
		if	os.path.isdir(self.tg1_d_mss1_c_badgood):			rmtree(self.tg1_d_mss1_c_badgood)
		if	os.path.isdir(self.tg1_d_mss1_c_goodbad):			rmtree(self.tg1_d_mss1_c_goodbad)
		if	not os.path.isfile(self.f_msb1_flower):				self.fmake(self.f_msb1_flower)

		# First sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_md))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_rock))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_scissors))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_md_gc))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_gc_sadmovies))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_gc_dumbasses))
		self.assertFalse(os.path.isfile(self.tg1_f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_bad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notbad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_notnot))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badbad))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_badgood))
		self.assertFalse(os.path.isdir(self.tg1_d_mss1_c_goodbad))
		self.assertTrue(os.path.isfile(self.f_msb1_flower))


		with self.assertLogs("Intergrowth", 20) as case_loggy:

			self.test_case = Forest()
			self.test_case.walk()

		self.assertCountEqual(
			case_loggy.output,
			[
				f"INFO:Intergrowth:Thrived twig \"{self.tg1_d_mss1_md_gc}\"",
				f"INFO:Intergrowth:Grown leaf \"{self.tg1_f_mss1_md_rock}\"",
				f"INFO:Intergrowth:Grown leaf \"{self.tg1_f_mss1_md_paper}\"",
				f"INFO:Intergrowth:Thrived twig \"{self.tg1_d_mss1_c_notnot}\"",
				f"INFO:Intergrowth:Thrived twig \"{self.tg1_d_mss1_c_bad}\"",
				f"INFO:Intergrowth:Thrived twig \"{self.tg1_d_mss1_c_notbad}\"",
				f"INFO:Intergrowth:Thrived twig \"{self.tg1_d_mss1_c_badbad}\"",
				f"INFO:Intergrowth:Thrived twig \"{self.tg1_d_mss1_c_badgood}\"",
				f"INFO:Intergrowth:Thrived twig \"{self.tg1_d_mss1_c_goodbad}\"",
				f"INFO:Intergrowth:Grown leaf \"{self.tg1_f_mss1_md_gc_dumbasses}\"",
				f"INFO:Intergrowth:Grown leaf \"{self.tg1_f_mss1_md_gc_sadmovies}\"",
				f"INFO:Intergrowth:Grown leaf \"{self.tg1_f_mss1_md_gc_onion}\"",
				f"INFO:Intergrowth:Grown leaf \"{self.mss1_flower}\"",
				f"INFO:Intergrowth:Thrived twig \"{self.d_mss1_c_good}\"",
				f"INFO:Intergrowth:Thrived twig \"{self.d_mss1_c_notgood}\"",
			]
		)


		# First sprout
		self.assertTrue(os.path.isfile(self.f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.f_mss1_md_scissors))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.d_mss1_c_goodbad))


		# First sprout first bough
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_md))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_rock))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_paper))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_scissors))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_md_gc))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_sadmovies))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_dumbasses))
		self.assertTrue(os.path.isfile(self.tg1_f_mss1_md_gc_onion))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_good))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notgood))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_bad))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notbad))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_notnot))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_badbad))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_badgood))
		self.assertTrue(os.path.isdir(self.tg1_d_mss1_c_goodbad))
		self.assertTrue(os.path.isfile(self.f_msb1_flower))








if	__name__ == "__main__" : unittest.main(verbosity=2)










import	os
import	unittest
from	pathlib							import Path
from	shutil							import rmtree
from	pygwarts.tests.hagrid			import HagridTestCase
from	pygwarts.irma.contrib			import LibraryContrib
from	pygwarts.hagrid.thrivables		import Tree
from	pygwarts.hagrid.thrivables		import Copse
from	pygwarts.hagrid.sprouts			import fssprout
from	pygwarts.hagrid.planting		import Flourish
from	pygwarts.hagrid.bloom.twigs		import Germination
from	pygwarts.hagrid.bloom.leafs		import Rejuvenation
from	pygwarts.hagrid.planting.leafs	import LeafGrowth
from	pygwarts.hagrid.planting.twigs	import TwigThrive








class InitCase(HagridTestCase):

	"""
		Some initiation cases.
	"""

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.INIT_HANDLER): os.remove(cls.INIT_HANDLER)

		if	os.path.isdir(cls.INIT_SET_BOUGH_1):	rmtree(cls.INIT_SET_BOUGH_1)
		if	os.path.isdir(cls.INIT_SET_BOUGH_2):	rmtree(cls.INIT_SET_BOUGH_2)
		if	os.path.isdir(cls.INIT_SET_BOUGH_3):	rmtree(cls.INIT_SET_BOUGH_3)
		if	os.path.isdir(cls.INIT_SET_SPROUT):		rmtree(cls.INIT_SET_SPROUT)

	@classmethod
	def setUpClass(cls):

		cls.make_loggy_file(cls, cls.INIT_HANDLER)
		if	os.path.isdir(cls.INIT_SET_BOUGH_1):		rmtree(cls.INIT_SET_BOUGH_1)
		if	os.path.isdir(cls.INIT_SET_BOUGH_2):		rmtree(cls.INIT_SET_BOUGH_2)
		if	not os.path.isdir(cls.INIT_SET_SPROUT):		os.makedirs(cls.INIT_SET_SPROUT)
		if	not os.path.isdir(cls.INIT_SET_BOUGH_3):	os.makedirs(cls.INIT_SET_BOUGH_3)




	def test_no_bough_tree(self):
		class Sakura(Tree):

			class loggy(LibraryContrib):

				handler		= self.INIT_HANDLER
				init_name	= "no_bough_tree"
				init_level	= 10

			class thrive(TwigThrive):	pass
			class folders(Germination):	pass

			class grow(LeafGrowth):		pass
			class files(Rejuvenation):	pass

			@fssprout(self.INIT_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("no_bough_tree", 10) as case_loggy:

			self.test_case = Sakura()
			self.assertIn("DEBUG:no_bough_tree:Sakura tree affected", case_loggy.output)
			self.assertIn(

				"DEBUG:no_bough_tree:Spreading complete, number of affected trees: 1",
				case_loggy.output
			)
			self.assertIn(

				"ERROR:no_bough_tree:Tree Sakura can't thrive due to "
				"AttributeError: Attribute 'bough' has no escalation to Sakura",
				case_loggy.output
			)
			self.assertIn("DEBUG:no_bough_tree:Discarded", case_loggy.output)
			self.assertEqual(len(self.test_case),0)








	def test_invalid_bough_tree(self):
		class Sakura(Tree):

			bough	= 42
			class loggy(LibraryContrib):

				handler		= self.INIT_HANDLER
				init_name	= "invalid_bough_tree"
				init_level	= 10

			class thrive(TwigThrive):	pass
			class folders(Germination):	pass

			class grow(LeafGrowth):		pass
			class files(Rejuvenation):	pass

			@fssprout(self.INIT_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("invalid_bough_tree", 10) as case_loggy:

			self.test_case = Sakura()
			self.assertIn("DEBUG:invalid_bough_tree:Sakura tree affected", case_loggy.output)
			self.assertIn(

				"DEBUG:invalid_bough_tree:Spreading complete, number of affected trees: 1",
				case_loggy.output
			)
			self.assertIn(f"INFO:invalid_bough_tree:Bough \"42\" is invalid", case_loggy.output)
			self.assertIn("DEBUG:invalid_bough_tree:Discarded", case_loggy.output)
			self.assertEqual(len(self.test_case),0)








	def test_uncreatable_bough_tree(self):

		# It was hard and unsuccessfull work to find unwritable by default folder in windows10
		if os.name == "posix":

			class Sakura(Tree):

				# The path which hagrid won't be able to create, likely cause of permissions.
				bough	= "/mnt/notcreation"
				class loggy(LibraryContrib):

					handler		= self.INIT_HANDLER
					init_name	= "uncreatable_bough_tree"
					init_level	= 10
				strict_thrive	= False

				class thrive(TwigThrive):	pass
				class folders(Germination):	pass

				class grow(LeafGrowth):		pass
				class files(Rejuvenation):	pass

				@fssprout(self.INIT_SET_SPROUT)
				class sync(Flourish):		pass


			with self.assertLogs("uncreatable_bough_tree", 10) as case_loggy:

				self.test_case = Sakura()
				self.assertIn("DEBUG:uncreatable_bough_tree:Sakura tree affected", case_loggy.output)
				self.assertIn(

					"DEBUG:uncreatable_bough_tree:Spreading complete, number of affected trees: 1",
					case_loggy.output
				)
				# Perhaps on different systems PermissionError raise may differ...
				self.assertIn(

					"ERROR:uncreatable_bough_tree:"
					f"Tree {self.test_case} can't thrive due to "
					f"PermissionError: [Errno 13] Permission denied: '{self.test_case.bough}'",
					case_loggy.output
				)
				self.assertIn("DEBUG:uncreatable_bough_tree:Discarded", case_loggy.output)
				self.assertEqual(len(self.test_case),0)








	def test_unwritable_bough_tree(self):

		# It was hard and unsuccessfull work to find unwritable by default folder in windows10
		if os.name == "posix":

			class Sakura(Tree):

				# Existent directory in which we cannot create folders/files
				bough	= "/mnt"
				class loggy(LibraryContrib):

					handler		= self.INIT_HANDLER
					init_name	= "unwritable_bough_tree"
					init_level	= 10

				class thrive(TwigThrive):	pass
				class folders(Germination):	pass

				class grow(LeafGrowth):		pass
				class files(Rejuvenation):	pass

				@fssprout(self.INIT_SET_SPROUT)
				class sync(Flourish):		pass


			with self.assertLogs("unwritable_bough_tree", 10) as case_loggy:

				self.test_case = Sakura()
				self.assertIn("DEBUG:unwritable_bough_tree:Sakura tree affected", case_loggy.output)
				self.assertIn(

					"DEBUG:unwritable_bough_tree:Spreading complete, number of affected trees: 1",
					case_loggy.output
				)
				self.assertIn(

					"ERROR:unwritable_bough_tree:"
					f"Tree {self.test_case} can't thrive due to "
					f"AssertionError: Bough \"{self.test_case.bough}\" thrive restricted",
					case_loggy.output
				)
				self.assertIn("DEBUG:unwritable_bough_tree:Discarded", case_loggy.output)
				self.assertEqual(len(self.test_case),0)








	def test_strict_thirve_tree(self):
		class Sakura(Tree):

			bough	= self.INIT_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.INIT_HANDLER
				init_name	= "strict_thirve_tree"
				init_level	= 10

			class thrive(TwigThrive):	pass
			class folders(Germination):	pass

			class grow(LeafGrowth):		pass
			class files(Rejuvenation):	pass

			@fssprout(self.INIT_SET_SPROUT)
			class sync(Flourish):		pass


		with self.assertLogs("strict_thirve_tree", 10) as case_loggy:

			self.test_case = Sakura()
			self.assertIn(

				"DEBUG:strict_thirve_tree:Sakura tree affected",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:strict_thirve_tree:Spreading complete, number of affected trees: 1",
				case_loggy.output
			)
			self.assertIn(
				f"INFO:strict_thirve_tree:Bough \"{self.INIT_SET_BOUGH_1}\" is invalid", case_loggy.output
			)
			self.assertIn("DEBUG:strict_thirve_tree:Discarded", case_loggy.output)
			self.assertEqual(len(self.test_case),0)








	def test_no_strict_thirve_tree(self):
		class Sakura(Tree):

			bough	= self.INIT_SET_BOUGH_1
			class loggy(LibraryContrib):

				handler		= self.INIT_HANDLER
				init_name	= "no_strict_thirve_tree"
				init_level	= 10
			strict_thrive	= False

			class thrive(TwigThrive):	pass
			class folders(Germination):	pass

			class grow(LeafGrowth):		pass
			class files(Rejuvenation):	pass

			@fssprout(self.INIT_SET_SPROUT)
			class sync(Flourish):		pass


		if	os.path.isdir(self.INIT_SET_BOUGH_1):	rmtree(self.INIT_SET_BOUGH_1)
		self.assertFalse(os.path.isdir(self.INIT_SET_BOUGH_1))


		with self.assertLogs("no_strict_thirve_tree", 10) as case_loggy:

			self.test_case = Sakura()
			self.assertIn("DEBUG:no_strict_thirve_tree:Sakura tree affected", case_loggy.output)
			self.assertIn(

				"DEBUG:no_strict_thirve_tree:Spreading complete, number of affected trees: 1",
				case_loggy.output
			)
			self.assertIn(f"DEBUG:no_strict_thirve_tree:Thriving single tree", case_loggy.output)
			self.assertEqual(len(self.test_case),2)


		if	os.path.isdir(self.INIT_SET_BOUGH_1):	rmtree(self.INIT_SET_BOUGH_1)
		self.assertFalse(os.path.isdir(self.INIT_SET_BOUGH_1))








	def test_no_bough_copse(self):
		class Grove(Copse):

			class loggy(LibraryContrib):

				handler		= self.INIT_HANDLER
				init_name	= "no_bough_copse"
				init_level	= 10

			class Sakura(Tree):			pass
			class Depth(Copse):
				class Sequoyah(Tree):	bough = self.INIT_SET_BOUGH_2
				class Hibiscus(Tree):	bough = self.INIT_SET_BOUGH_3

			class thrive(TwigThrive):	pass
			class folders(Germination):	pass

			class grow(LeafGrowth):		pass
			class files(Rejuvenation):	pass

			@fssprout(self.INIT_SET_SPROUT)
			class sync(Flourish):		pass


		if	not os.path.isdir(self.INIT_SET_BOUGH_3):	os.makedirs(self.INIT_SET_BOUGH_3)
		self.assertTrue(os.path.isdir(self.INIT_SET_BOUGH_3))

		with self.assertLogs("no_bough_copse", 10) as case_loggy:

			self.test_case = Grove()
			self.assertNotIn(

				"DEBUG:no_bough_copse:Grove.Sakura tree affected",
				case_loggy.output
			)
			self.assertNotIn(

				"DEBUG:no_bough_copse:Grove.Depth.Sequoyah tree affected",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:no_bough_copse:Grove.Depth.Hibiscus tree affected",
				case_loggy.output
			)
			self.assertIn(

				"DEBUG:no_bough_copse:Spreading complete, number of affected trees: 1",
				case_loggy.output
			)
			self.assertIn(
				f"INFO:no_bough_copse:Bough \"{self.INIT_SET_BOUGH_2}\" is invalid", case_loggy.output
			)
			self.assertIn(

				"ERROR:no_bough_copse:Tree Grove.Sakura can't thrive due to "
				f"AttributeError: Attribute 'bough' has no escalation to Grove",
				case_loggy.output
			)
			self.assertEqual(len(self.test_case.Sakura),0)
			self.assertEqual(len(self.test_case.Depth.Sequoyah),0)
			self.assertEqual(len(self.test_case.Depth.Hibiscus),2)








	def test_bad_sprouts(self):
		class Sakura(Tree):

			bough	= self.INIT_SET_BOUGH_3
			class loggy(LibraryContrib):

				handler		= self.INIT_HANDLER
				init_name	= "bad_sprouts"
				init_level	= 10

			@fssprout(Path(self.INIT_SET_BOUGH_1))
			@fssprout(self.INIT_SET_BOUGH_2)
			@fssprout(self.INIT_SET_SPROUT)
			class sync(Flourish):	pass


		with self.assertLogs("bad_sprouts", 10) as case_loggy:

			self.test_case = Sakura()
			self.test_case.sync()


		self.assertIn(
			f"CRITICAL:bad_sprouts:Sprout \"{self.INIT_SET_BOUGH_1}\" is invalid", case_loggy.output
		)
		self.assertIn(
			f"CRITICAL:bad_sprouts:Sprout \"{self.INIT_SET_BOUGH_2}\" is not thrived", case_loggy.output
		)








if	__name__ == "__main__" : unittest.main(verbosity=2)










from __future__								import annotations
from shelve									import open		as shopen
from os										import access	as osaccess
from os										import path		as ospath
from os										import R_OK
from os										import makedirs
from typing									import Hashable
from typing									import Any
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.magical.chests				import KeyChest
from pygwarts.magical.spells				import patronus
from pygwarts.magical.spells				import flagrate








class LibraryShelf(Transmutable):

	"""
		irma core class, that represents the wrapper over standard library "shelve" functionality, that
		integrated in mutable chain. Such integration allows not only maintain the shelf object as
		dictionary, with ability to store and load it as files, but also maintain another one, auxiliary
		shelf object as dictionary. Those shelf are:
			"real_shelf"	- the "main" dictionary for LibraryShelf object, which means it is the
							LibraryShelf object. Every method or property (except __call__) operates
							with "real_shelf". When "grab" method works it populates exactly "real_shelf".
			"magical_shelf"	- the shelf which serves as some sort of buffer for "real_shelf". Operations
							via __setitem__ and __delitem__ works for both shelf, setting and deleting
							mapping in both "real_shelf" and "magical_shelf", correspondingly. The only
							method __call__ works only for "magical_shelf" and serves as a way to operate
							with it the way it is possible to operate with "real_shelf" with LibraryShelf
							members.
		As both "real_shelf" and "magical_shelf" are KeyChest objects, all operations on them might be
		considered identical. The main idea of maintaining "magical_shelf" is that when "real_shelf"
		being populated, the further operations might all be reflected in "magical_shelf", so it will
		be the newer version of "real_shelf". It is recommended to use only defined members to operate
		with LibraryShelf object, and not direct access to "real_shelf" and "magical_shelf", cause
		LibraryShelf object changes are tracked by "modified" field, which reacts to the events that are
		actually modifies LibraryShelf object. Direct access to "real_shelf" will not alter the "modified"
		state properly, so "produce" method might not work and shelf will not be produced as file.
		Main helper methods to maintain input and output shelf files:
			grab	- opens shelf file as a dictionary and puts it's key-value pairs to the "real_shelf".
			produce	- opens shelf file as a dictionary and puts "real_shelf" or "magical_shelf" key-value
					pairs to that file, so it will be saved as a file.
		Fields that helps maintain input and output shelf files.
			grabbing	- input file absolute path to extract shelf from
			producing	- output file absolute path to save current shelf dictionary
			reclaiming	- boolean flag that indicates whether or not use input file absolute path
						("grabbing") as path for output file. If "producing" provided too, it will
						be ignored.
	"""


	grabbing	:str	=None
	producing	:str	=None
	reclaiming	:bool	=False


	class real_shelf(KeyChest):		pass
	class magical_shelf(KeyChest):	pass


	def __hash__(self) -> int:	return id(self)
	def __len__(self) -> int :	return len(self.real_shelf)
	def __contains__(self, K :Hashable)	-> bool			: return K in self.real_shelf
	def __getitem__(self, K :Hashable)	-> Any | None	: return self.real_shelf[K]
	def __setitem__(self, K :Hashable, V :Any):

		"""
			Assigning for both "real_shelf" and "magical_shelf".
			The "modified" flag will be set to True if all following conditions are met:
				- either key was not present in "real_shelf" or mapped value is different;
				- key-value successfully mapped for "real_shelf";
				- key-value successfully mapped for "magical_shelf";
		"""

		presence	= K in self.real_shelf
		valueness	= self.real_shelf[K] == V


		real = self.real_shelf(K,V)
		magical = self.magical_shelf(K,V)


		if(
			(not presence or not valueness)
			and real is not None
			and magical is not None
		):	self.modified = True




	def __delitem__(self, K :Hashable):

		""" Delete for both "real_shelf" and "magical_shelf". """

		current = len(self.real_shelf)


		del self.real_shelf[K]
		del self.magical_shelf[K]
		if	len(self.real_shelf) <current : self.modified = True




	def __eq__(self, other :LibraryShelf | KeyChest | dict) -> bool :

		"""
			Comparison of equality of current Shelf "real_shelf" with other Shelf "real_shelf",
			or another KeyChest or just a dict object.
		"""

		match other:

			case LibraryShelf()		: return 	self.real_shelf == other.real_shelf
			case KeyChest() | dict(): return 	self.real_shelf == other
			case _					: return 	False




	def __gt__(self, other :LibraryShelf | KeyChest | dict) -> bool | TypeError :

		"""
			Comparison of greatness of current Shelf "real_shelf" with other Shelf "real_shelf",
			or another KeyChest or just a dict object.
		"""

		match other:

			case LibraryShelf()		: return	len(self.real_shelf) > len(other.real_shelf)
			case KeyChest() | dict(): return	len(self.real_shelf) > len(other)
			case _					: raise		TypeError(f"Object \"{other}\" cannot be compared with {self}")




	def __ge__(self, other :LibraryShelf | KeyChest | dict) -> bool | TypeError :

		"""
			Comparison of greatness or equality of current Shelf "real_shelf" with other Shelf "real_shelf",
			or another KeyChest or just a dict object.
		"""

		match other:

			case LibraryShelf()		: return	len(self.real_shelf) >= len(other.real_shelf)
			case KeyChest() | dict(): return	len(self.real_shelf) >= len(other)
			case _					: raise		TypeError(f"Object \"{other}\" cannot be compared with {self}")




	def __lt__(self, other :LibraryShelf | KeyChest | dict) -> bool | TypeError :

		"""
			Comparison of lesserness of current Shelf "real_shelf" with other Shelf "real_shelf",
			or another KeyChest or just a dict object.
		"""

		match other:

			case LibraryShelf()		: return	len(self.real_shelf) < len(other.real_shelf)
			case KeyChest() | dict(): return	len(self.real_shelf) < len(other)
			case _					: raise		TypeError(f"Object \"{other}\" cannot be compared with {self}")




	def __le__(self, other :LibraryShelf | KeyChest | dict) -> bool | TypeError :

		"""
			Comparison of lesserness or equality of current Shelf "real_shelf" with other Shelf "real_shelf",
			or another KeyChest or just a dict object.
		"""

		match other:

			case LibraryShelf()		: return	len(self.real_shelf) <= len(other.real_shelf)
			case KeyChest() | dict(): return	len(self.real_shelf) <= len(other)
			case _					: raise		TypeError(f"Object \"{other}\" cannot be compared with {self}")




	# Any iteration for current Shelf occurs along "real_shelf", so it's dunders just delegated to it,
	# implementing the very same way it should with the "real_shelf" KeyChest.
	def __iter__(self)		: yield		from self.real_shelf
	def __next__(self)		: return	next(self.real_shelf)
	def __reversed__(self)	: yield		from reversed(self.real_shelf)




	def __call__(self, *load, silent :bool =False) -> Any :

		"""
			Maintains "magical_shelf" functionality. Providing one argument will be treated as indexing
			to "magical_shelf" so the mapped value will be returned. Providing two arguments as key-value
			pair will lead to populating shelf. Argument "silent" allow to not trigger "modified" flag
			when putting. No arguments provides direct access to "magical_shelf".
		"""

		match load:

			case ( K, )	:	return	self.magical_shelf[K]
			case ( K,V ):

				presence	= K in self.magical_shelf
				valueness	= self.magical_shelf[K] == V
				magical		= self.magical_shelf(K,V)


				if	(not presence or not valueness) and magical is not None:

					if		not silent: self.modified = True
					else:	self.magical_shelf.loggy.debug("Silent mode operational")


			case ()	: return self.magical_shelf
			case _	: self.loggy.warning(f"Unable to put \"{load}\" in magical Shelf")




	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)


		self.modified = False
		if	self.grab() is None : self.loggy.debug("New Shelf successfully initiated")




	# Property that calculates and returns current difference between "real_shelf" and "magical_shelf"
	# lengths. As calculation suggest subtraction of "magical_shelf" length from "real_shelf", the
	# results must be interpreted as wollows:
	# 0		- shelves have same lengths;
	# N		- "real_shelf" bigger than "magical_shelf" by N items;
	# -N	- "magical_shelf" bigger than "real_shelf" by N items;
	# Also properties that returns the set of keys which are differs for real to magical or vice versa.
	@property
	def diff(self) -> int : return len(self.real_shelf) - len(self.magical_shelf)
	@property
	def real_diff(self) -> Set[Hashable]	: return set(self.keys()) - set(self.magical_shelf.keys())
	@property
	def magical_diff(self) -> Set[Hashable]	: return set(self.magical_shelf.keys()) - set(self.keys())




	# Utility methods that are redirections to "real_shelf" implementations of "Inspection",
	# "keysof" and "keys". Must work absolutely the same way as "real_shelf" implies.
	def Inspection(self, level :int =None) -> str	: return self.real_shelf.Inspection(level)
	def keysof(self, value :Any) -> List[Hashable]	: return self.real_shelf.keysof(value)
	def keys(self) -> List[Hashable]				: return self.real_shelf.keys()




	def grab(	self,
				grabbing_path	:str	=None,
				*,
				rewrite			:bool	=False,
				from_locker		:bool	=True
			)-> str | None		:

		"""
			Helper method that allows to populate "real_shelf" with key-value pairs from a shelve file.
			Path to the file might be provided by an argument, or will be taken from "grabbing" field,
			and must be an existent shelve file. Argument "rewrite" allows to preliminary clear current
			"real_shelf". The "from_locker" flag decides whether to iterate through special _locker_ key
			in shelved file, which might be produced by LibraryShelf, and use it for indexing shelved file,
			hence the order of "real_shelf" populating will be the same. By default it is True, cause
			there is no difference for final "real_shelf" length whether to insert values in "_inside_" and
			"_locker_" of "real_shelf" KeyChest in, probably found another "_locker_", order or not.
			If invoked within initiation, using grabbing field, "modified" flag not toggled to True.
			If invoked as a method with "grabbing_path" argument, "modified" flag considered to be True
			if "real_shelf" length have changed. The logic behind such particular behavior is that when
			LibraryShelf is initiated it either an empty or grabbed from file, which is the same in terms
			of "modified" flag. Even regrabbing from "grabbing" field doesn't lead to "modified" flag
			toggling, cause the LibraryShelf object might be already "modified". Only grabbing with
			argument provided will change "modified" flag. Return grabbed file path in case of success,
			None otherwise.
		"""

		if	(grabbable := grabbing_path or self.grabbing) is not None:
			if	isinstance(grabbable, str):
				if(
					(
						ospath.isfile(grabbable)
						or
						ospath.isfile(grabbable + ".db")
						or
						ospath.isfile(grabbable + ".dat")
					)
					and
					(
						osaccess(grabbable, R_OK)
						or
						osaccess(grabbable + ".db", R_OK)
						or
						osaccess(grabbable + ".dat", R_OK)
					)
				):


					self.loggy.debug(f"Grabbing Shelf \"{grabbable}\"")
					L = len(self.real_shelf)


					try:

						with shopen(grabbable) as outer:
							self.loggy.debug(f"Shelf size {len(outer)} keys")


							if	rewrite: self.real_shelf.unload()


							# As KeyChests maintains "_locker_" attribute as a list of keys to preserve
							# the order keys were added to chest, such an attribute will be assigned to
							# the Shelf by "produce" method. By obtaining this, grabbing will occurs
							# along keys in it. In case "_locker_" couldn't be find, order might break.
							if	from_locker and (locker := outer.get("_locker_")) is not None:
								self.loggy.debug(f"Obtained locker of length {len(locker)}")


								for k in locker : self.real_shelf(k, outer[k])


							else:
								if from_locker: self.loggy.debug("Locker not obtained, order not granted")
								else: self.loggy.debug("Locker skipped, order not granted")


								for k,v in outer.items():
									if	k != "_locker_" : self.real_shelf(k,v)


							self.loggy.debug(f"Shelf \"{grabbable}\" successfully grabbed")


							# The point of decision on whether LibraryShelf object might be considered
							# modified or not, based on whether "grab" was in __init__ time (not modified)
							# or by provided "grabbing_path" argument (modified).
							if		grabbing_path is not None : self.modified = len(self.real_shelf) != L
							return	grabbable


					except Exception as E:


						self.loggy.warning(f"Failed to grab \"{grabbable}\" due to: {patronus(E)}")
				else:	self.loggy.info(f"Shelf \"{grabbable}\" to grab does not exist or inaccessible")
			else:		self.loggy.warning(f"Invalid grabbing path type \"{type(grabbable)}\"")
		else:			self.loggy.debug("Shelf to grab not provided")




	def produce(	self,
					producing_path	:str	=None,
					*,
					rewrite			:bool	=False,
					magical			:bool	=False,
					ignore_mod		:bool	=False,
					strict_mode		:bool	=True,
					locker_mode		:bool	=False,
				)-> str | None		:

		"""
			Helper method that produces current object "real_shelf" or "magical_shelf" to a shelve file.
			Path to the file might be provided by an argument, or will be derived from either "producing"
			field or from grabbing field, in case of "reclaiming" flag set to True ("producing" has
			priority over "reclaiming"). With argument "rewrite" it is possible to fully clear producible
			shelve file, as it is opened as a new or as already existent file. Decision of which shelf to
			produce is made by flag "magical", which is False by default to produce "real_shelf".
			Argument "ignore_mod" allows to produce even if there was no Shelfs modification, described by
			falsy flag "modified". With "strict_mode" set to True it is assumed, that file destination
			is in existent folder, and Exception will be caught otherwise, but if False, check-creation
			will be granted. The "locker_mode" is disabled by default, cause it leads to end up mapping
			changes by LibraryShelf, so such feature must be turned on with full understanding of further
			consequences, pros and cons of one more key "_locker_" in mapping. Returns produced file path
			in case of success, None otherwise.
		"""

		if	self.modified or ignore_mod:
			if(

				producible := producing_path
				or
				self.producing
				or
				(self.grabbing if self.reclaiming else None)

			)	is not None:

				shl = len(self.magical_shelf if magical else self.real_shelf)
				self.loggy.debug(f"Producing Shelf \"{producible}\"")
				self.loggy.debug("Source is %s Shelf with %s key%s"%(
					"magical" if magical else "real", shl, flagrate(shl)
				))

				try:

					# Boolean flag "strict_mode" disallow creating non-existing parents directories
					# for Shelf file producing, by default. Toggled to False allows check-creation.
					if	not strict_mode : makedirs(ospath.dirname(producible), exist_ok=True)
					if	ignore_mod : self.loggy.debug("Shelf modification flag ignored")


					with shopen(producible, "n" if rewrite else "c") as outer:
						for k,v in self.magical_shelf if magical else self.real_shelf:


							outer[k] = v
							self.loggy.debug(f"Produced key \"{k}\"")


						# Saving current procssed Shelf "_locker_" as a special key in the producible,
						# to allow further grabbing keep the order it was in KeyChest.
						if	locker_mode:

							locker = (self.magical_shelf if magical else self.real_shelf)._locker_
							outer["_locker_"] = locker
							self.loggy.debug(f"Locker of length {len(locker)} produced")


						if rewrite:	self.loggy.info(f"Rewritten Shelf \"{producible}\"")
						else:		self.loggy.info(f"Produced Shelf \"{producible}\"")


						return	producible


				except	Exception as E:


					self.loggy.warning(f"Producing \"{producible}\" failed with {patronus(E)}")
			else:	self.loggy.debug("Shelf to produce not provided")
		else:		self.loggy.debug("Shelf was not modified")




	def unload(self, magical :bool =False):

		"""
			Executes both real and magical KeyChests "unload" operation, that completely clears both
			KeyChests key-pairs and lockers, and reset their corresponding counters. Also set "modified"
			flag to False, meaning full initiation imitation.
		"""

		self.modified = False
		self.real_shelf.unload()
		self.loggy.info(f"{self} real shelf was unloaded")

		if	magical:

			self.magical_shelf.unload()
			self.loggy.info(f"{self} magical shelf was unloaded")







